# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for measurement results."""

import numpy as np
import pytest

from mitiq import MeasurementResult


@pytest.mark.parametrize("asarray", (True, False))
@pytest.mark.parametrize("qubit_indices", ((0, 1), (1, 20)))
def test_measurement_result(asarray, qubit_indices):
    bitstrings = [[0, 0], [0, 1], [1, 0]]
    if asarray:
        bitstrings = np.array(bitstrings)
    result = MeasurementResult(bitstrings, qubit_indices=qubit_indices)

    assert result.nqubits == 2
    assert result.qubit_indices == qubit_indices
    assert result.shots == 3
    assert np.allclose(result.result, bitstrings)


def test_measurement_result_bad_qubit_indices():
    with pytest.raises(ValueError, match="MeasurementResult has"):
        MeasurementResult([[0], [1]], qubit_indices=(1, 5))


def test_measurement_result_not_bits():
    with pytest.raises(ValueError, match="should look like"):
        MeasurementResult(result=[[2]])

    with pytest.raises(ValueError, match="should look like"):
        MeasurementResult(result=[[0, 0], [0, 1], [-1, 0]])


def test_measurement_result_invoked_with_dict():
    with pytest.raises(TypeError, match="from_counts"):
        MeasurementResult({"001": 123, "010": 456})


def test_filter_qubits():
    result = MeasurementResult([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
    assert np.allclose(result.filter_qubits([0]), np.array([[0], [0], [1]]))
    assert np.allclose(result.filter_qubits([1]), np.array([[0], [1], [0]]))
    assert np.allclose(result.filter_qubits([2]), np.array([[1], [0], [0]]))

    assert np.allclose(
        result.filter_qubits([0, 1]), np.array([[0, 0], [0, 1], [1, 0]])
    )
    assert np.allclose(
        result.filter_qubits([0, 2]), np.array([[0, 1], [0, 0], [1, 0]])
    )
    assert np.allclose(
        result.filter_qubits([1, 2]), np.array([[0, 1], [1, 0], [0, 0]])
    )


def test_empty():
    result = MeasurementResult([])
    assert result.nqubits == 0
    assert result.shots == 0
    assert result.result == []


def test_convert_to_array():
    bitstrings = [[0, 0], [0, 1], [1, 0]]
    result = MeasurementResult(bitstrings)
    assert np.allclose(result.asarray, np.array(bitstrings))

    assert np.allclose(MeasurementResult([]).asarray, np.array([]))


@pytest.mark.parametrize("qubit_indices", ((0, 1), (1, 20)))
def test_measurement_result_with_strings(qubit_indices):
    """Try using strings instead of lists of integers."""
    bitstrings = ["00", "01", "10"]
    int_bitstrings = [[0, 0], [0, 1], [1, 0]]

    result = MeasurementResult(bitstrings, qubit_indices=qubit_indices)

    assert result.nqubits == 2
    assert result.qubit_indices == qubit_indices
    assert result.shots == 3
    assert result.result == int_bitstrings


@pytest.mark.parametrize("qubit_indices", ((0, 1), (1, 20)))
def test_measurement_result_from_counts(qubit_indices):
    """Test initialization from a dictionary of counts."""
    counts = {"00": 1, "01": 2}
    int_bitstrings = [[0, 0], [0, 1], [0, 1]]

    result = MeasurementResult.from_counts(
        counts=counts,
        qubit_indices=qubit_indices,
    )
    assert result.nqubits == 2
    assert result.qubit_indices == qubit_indices
    assert result.shots == 3
    assert result.result == int_bitstrings


def test_measurement_result_get_counts():
    """Test initialization from a dictionary of counts."""
    counts = {"00": 1, "01": 2}

    int_bitstrings = [[0, 0], [0, 1], [0, 1]]
    result = MeasurementResult(
        result=int_bitstrings,
        qubit_indices=(1, 20),
    )
    assert result.get_counts() == counts
    # Info about qubit indices is expected to be lost
    new_res = MeasurementResult.from_counts(result.get_counts())
    assert new_res.qubit_indices == (0, 1)  # Default values


def test_measurement_result_to_from_dictionary():
    """Test initialization from a dictionary of counts."""
    data = {
        "counts": {"00": 1, "01": 2},
        "shots": 3,
        "nqubits": 2,
        "qubit_indices": (1, 7),
    }
    assert MeasurementResult.from_dict(data).to_dict() == data


def test_measurement_repr_():
    """Test string representation and printing."""
    counts = {"00": 1000, "01": 200}
    result = MeasurementResult.from_counts(
        counts=counts,
        qubit_indices=(1, 20),
    )
    expected = (
        "MeasurementResult: {'nqubits': 2, 'qubit_indices': (1, 20),"
        " 'shots': 1200, 'counts': {'00': 1000, '01': 200}}"
    )
    assert repr(result) == expected
    assert str(result) == expected


def test_measurement_result_prob_distribution():
    """Test initialization from a dictionary of counts."""
    int_bitstrings = [[0, 0], [0, 1], [0, 1], [0, 1]]
    result = MeasurementResult(
        result=int_bitstrings,
        qubit_indices=(1, 20),
    )
    dist = result.prob_distribution()
    assert len(dist) == 2
    assert np.isclose(dist["00"], 0.25)
    assert np.isclose(dist["01"], 0.75)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mirror circuits."""

import cirq
import networkx as nx
import pytest
from numpy import random

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks import mirror_circuits
from mitiq.utils import _equal

paulis = mirror_circuits.paulis
cliffords = mirror_circuits.cliffords
single_cliffords = []
for c in cliffords:
    single_cliffords.extend(c)
all_cliffords = single_cliffords.copy()
all_cliffords.append(cirq.CNOT)
all_gates = paulis + all_cliffords
all_gates.append(cirq.measure)


@pytest.mark.parametrize(
    "graph",
    [
        nx.Graph({3: (4,), 4: (3, 6), 5: (6,)}),
        nx.Graph({0: (99,), 10: (20,), 5: (6,)}),
    ],
)
def test_random_paulis(graph):
    circuit = mirror_circuits.random_paulis(
        connectivity_graph=graph, random_state=random.RandomState()
    )
    assert isinstance(circuit, cirq.Circuit)
    assert len(circuit.all_qubits()) == len(graph.nodes)
    for qubit in circuit.all_qubits():
        assert qubit.x in graph.nodes
    assert len(list(circuit.all_operations())) == len(graph.nodes)
    assert set(op.gate for op in circuit.all_operations()).issubset(paulis)


def test_edge_grab():
    nqubits = [9, 10, 11]
    graphs = []
    for qubits in nqubits:
        graphs.append(nx.complete_graph(qubits))
    edges_to_remove = [
        [(1, 2), (4, 7), (3, 5)],
        [(2, 4), (5, 6), (7, 8)],
        [(3, 4), (5, 6), (8, 9)],
    ]

    for x in range(3):
        graph = graphs[x]
        for edge in edges_to_remove[x]:
            graph.remove_edge(*edge)
        selected_edges = mirror_circuits.edge_grab(
            0.3, graph, random_state=random.RandomState()
        )
        assert set(selected_edges.edges()).issubset(graph.edges())
        assert selected_edges.number_of_nodes() == nqubits[x]
        for node in selected_edges.nodes:
            assert len(list(selected_edges.neighbors(node))) in (0, 1)


def test_random_cliffords():
    nqubits = [9, 10, 11]
    edges = [
        [(1, 2), (4, 7), (3, 5)],
        [(2, 4), (5, 6), (7, 8)],
        [(3, 4), (5, 6), (8, 9)],
    ]
    for x in range(3):
        graph = nx.Graph()
        graph.add_nodes_from(range(nqubits[x]))
        graph.add_edges_from(edges[x])
        circuit = mirror_circuits.random_cliffords(
            graph, random_state=random.RandomState()
        )
        assert isinstance(circuit, cirq.Circuit)
        assert len(circuit.all_qubits()) == nqubits[x]
        two_q_gates = set()
        for a, b in edges[x]:
            two_q_gates.add(cirq.CNOT(cirq.LineQubit(a), cirq.LineQubit(b)))
        assert two_q_gates.issubset(circuit.all_operations())
        assert set(op.gate for op in circuit.all_operations()).issubset(
            all_cliffords
        )


def test_random_single_cliffords():
    qubits = [[0, 1, 2, 3, 4], [1, 3, 5, 7], [1, 2, 4, 6, 7, 8]]
    for x in range(3):
        graph = nx.Graph()
        graph.add_nodes_from(qubits[x])
        circuit = mirror_circuits.random_single_cliffords(
            graph, random_state=random.RandomState()
        )
        assert isinstance(circuit, cirq.Circuit)
        assert set(circuit.all_qubits()).issubset
        ([cirq.LineQubit(qubit) for qubit in qubits[x]])
        assert set(op.gate for op in circuit.all_operations()).issubset(
            single_cliffords
        )


@pytest.mark.parametrize(
    "depth_twoqprob_graph",
    [
        (16, 0.3, nx.complete_graph(3)),
        (20, 0.4, nx.complete_graph(4)),
        (24, 0.5, nx.complete_graph(5)),
    ],
)
def test_generate_mirror_circuit(depth_twoqprob_graph):
    depth, xi, connectivity_graph = depth_twoqprob_graph
    n = connectivity_graph.number_of_nodes()
    circ, _ = mirror_circuits.generate_mirror_circuit(
        depth, xi, connectivity_graph
    )
    assert isinstance(circ, cirq.Circuit)
    assert len(circ.all_qubits()) == n
    assert set(op.gate for op in circ.all_operations()).issubset(all_gates)
    circ.append(cirq.measure(*cirq.LineQubit.range(n)))
    result = (
        cirq.Simulator()
        .run(circ, repetitions=1_000)
        .multi_measurement_histogram(keys=circ.all_measurement_key_names())
    )
    assert (
        len(result.keys()) == 1
    )  # checks that the circuit only outputs one bitstring


@pytest.mark.parametrize("seed", (0, 3))
def test_mirror_circuit_seeding(seed):
    nlayers = 5
    two_qubit_gate_prob = 0.4
    connectivity_graph = nx.complete_graph(5)
    circuit, _ = mirror_circuits.generate_mirror_circuit(
        nlayers, two_qubit_gate_prob, connectivity_graph, seed=seed
    )
    for _ in range(5):
        circ, _ = mirror_circuits.generate_mirror_circuit(
            nlayers, two_qubit_gate_prob, connectivity_graph, seed=seed
        )
        assert _equal(
            circuit,
            circ,
            require_qubit_equality=True,
            require_measurement_equality=True,
        )


@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_mirror_circuits_conversions(return_type):
    nlayers = 5
    two_qubit_gate_prob = 0.4
    connectivity_graph = nx.complete_graph(5)
    circuit, _ = mirror_circuits.generate_mirror_circuit(
        nlayers,
        two_qubit_gate_prob,
        connectivity_graph,
        return_type=return_type,
    )
    assert return_type in circuit.__module__


@pytest.mark.parametrize(
    "twoq_name_and_gate", [("CNOT", cirq.CNOT), ("CZ", cirq.CZ)]
)
def test_two_qubit_gate(twoq_name_and_gate):
    twoq_name, twoq_gate = twoq_name_and_gate
    circuit, _ = mirror_circuits.generate_mirror_circuit(
        nlayers=2,
        two_qubit_gate_prob=1.0,
        connectivity_graph=nx.complete_graph(5),
        two_qubit_gate_name=twoq_name,
    )
    two_qubit_gates = {
        op.gate for op in circuit.all_operations() if len(op.qubits) == 2
    }
    assert two_qubit_gates == {twoq_gate}


def test_two_qubit_gate_unsupported():
    with pytest.raises(ValueError, match="Supported two-qubit gate names are"):
        mirror_circuits.generate_mirror_circuit(
            1, 1.0, nx.complete_graph(2), two_qubit_gate_name="bad_gate_name"
        )


def test_two_qubit_gate_bad_probability():
    with pytest.raises(
        ValueError, match="two_qubit_gate_prob must be between 0 and 1"
    ):
        mirror_circuits.generate_mirror_circuit(1, 2.0, nx.complete_graph(2))


def test_deterministic_correct_bitstrings():
    """For a fixed seed, correct bitstrings should be deterministic."""
    expected_correct_bitstrings = (
        [[0, 0], [1, 1], [0, 0], [0, 1], [0, 0], [0, 0], [0, 0], [0, 0]]
        + [[0, 0], [1, 1], [1, 1], [0, 1], [0, 1], [0, 1], [1, 0], [0, 1]]
        + [[1, 1], [1, 0], [1, 0], [1, 0], [0, 0], [1, 1], [0, 0], [0, 1]]
    )
    for j, expected in enumerate(expected_correct_bitstrings):
        _, bitstring = mirror_circuits.generate_mirror_circuit(
            nlayers=1,
            two_qubit_gate_prob=1.0,
            connectivity_graph=nx.complete_graph(2),
            seed=j,
        )
        assert bitstring == expected



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for qiskit TransformationPass (transpiler.py)."""

import pytest
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.converters import circuit_to_dag
from qiskit.transpiler import Layout
from qiskit.transpiler.exceptions import TranspilerError

from mitiq.interface.mitiq_qiskit import ApplyMitiqLayout, ClearLayout


def test_apply_mitiq_layout_success():
    """Tests ApplyMitiqLayout successful run."""
    # Setup: Create a simple quantum circuit and convert it to a DAG
    qr = QuantumRegister(3, name="qr")
    circuit = QuantumCircuit(qr)
    circuit.h(qr[0])
    circuit.cx(qr[0], qr[2])
    dag = circuit_to_dag(circuit)

    new_qregs = [QuantumRegister(3, name="new_qr")]
    layout_pass = ApplyMitiqLayout(new_qregs=new_qregs)

    # Manually set a layout in the property_set (usually done by previous pass)
    layout_pass.property_set = {
        "layout": Layout.from_qubit_list(circuit.qubits)
    }

    # Run the transformation
    new_dag = layout_pass.run(dag)

    # Assertions: Verify the new DAG has the expected structure
    assert len(new_dag.qubits) == len(qr)
    for qubit in new_dag.qubits:
        assert qubit._register.name == "new_qr"


def test_apply_mitiq_layout_fail_small_register():
    """Tests ApplyMitiqLayout run fails with too small a QuantumRegister."""
    # Setup: Create a simple quantum circuit and convert it to a DAG
    qr = QuantumRegister(3, name="qr")
    circuit = QuantumCircuit(qr)
    circuit.h(qr[0])
    circuit.cx(qr[0], qr[2])
    dag = circuit_to_dag(circuit)

    new_qregs = [QuantumRegister(2, name="new_qr")]
    layout_pass = ApplyMitiqLayout(new_qregs=new_qregs)

    # Manually set a layout in the property_set (usually done by previous pass)
    layout_pass.property_set = {
        "layout": Layout.from_qubit_list(circuit.qubits)
    }

    # Expect TranspilerError due to small QuantumRegister
    with pytest.raises(TranspilerError):
        layout_pass.run(dag)


def test_apply_mitiq_layout_fail_no_layout():
    """Tests ApplyMitiqLayout run fails with no layout."""
    # Setup a basic circuit and DAG
    qr = QuantumRegister(2, name="qr")
    circuit = QuantumCircuit(qr)
    dag = circuit_to_dag(circuit)

    layout_pass = ApplyMitiqLayout(new_qregs=[QuantumRegister(2, "new_qr")])

    # Make sure there is no layout
    if layout_pass.property_set["layout"]:
        del layout_pass.property_set["layout"]

    # Expect TranspilerError due to missing layout
    with pytest.raises(TranspilerError):
        layout_pass.run(dag)


def test_clear_layout_clears_the_layout():
    """Tests ClearLayout clears the layout."""
    # Setup: Create a simple quantum circuit and convert it to a DAG
    qr = QuantumRegister(2, name="qr")
    circuit = QuantumCircuit(qr)
    dag = circuit_to_dag(circuit)

    layout_pass = ClearLayout()

    # Manually set a layout in the property_set (usually done by previous pass)
    layout_pass.property_set = {
        "layout": Layout.from_qubit_list(circuit.qubits)
    }

    layout_pass.run(dag)
    assert "layout" not in layout_pass.property_set



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for scaling noise by unitary folding of layers in the input
circuit to allow for multivariate extrapolation."""

import math
from copy import deepcopy

import pytest
from cirq import Circuit, LineQubit, ops

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.interface import convert_from_mitiq
from mitiq.lre.multivariate_scaling.layerwise_folding import (
    _get_chunks,
    _get_num_layers_without_measurements,
    get_scale_factor_vectors,
    multivariate_layer_scaling,
)

qreg1 = LineQubit.range(3)
test_circuit1 = Circuit(
    [ops.H.on_each(*qreg1)],
    [ops.CNOT.on(qreg1[0], qreg1[1])],
    [ops.X.on(qreg1[2])],
    [ops.TOFFOLI.on(*qreg1)],
)

test_circuit1_with_measurements = deepcopy(test_circuit1)
test_circuit1_with_measurements.append(ops.measure_each(*qreg1))


def test_multivariate_layerwise_scaling_num_circuits():
    """Ensure the correct number of circuits are generated."""
    degree, fold_multiplier, num_chunks = 2, 2, 3
    scaled_circuits = multivariate_layer_scaling(
        test_circuit1, degree, fold_multiplier, num_chunks
    )

    depth = len(test_circuit1)
    # number of circuit is `degree` + `depth` choose `degree`
    assert len(scaled_circuits) == math.comb(degree + depth, degree)


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES)
def test_multivariate_layerwise_scaling_types(circuit_type):
    """Ensure layer scaling returns circuits of the correct type."""
    circuit = convert_from_mitiq(test_circuit1, circuit_type.name)
    degree, fold_multiplier, num_chunks = 2, 2, 3
    scaled_circuits = multivariate_layer_scaling(
        circuit, degree, fold_multiplier, num_chunks
    )

    assert all(
        isinstance(circuit, circuit_type.value) for circuit in scaled_circuits
    )


def test_multivariate_layerwise_scaling_cirq():
    """Ensure the folding pattern is correct for a nontrivial case."""
    scaled_circuits = multivariate_layer_scaling(test_circuit1, 2, 2, 3)
    folding_pattern = [
        (1, 1, 1),
        (5, 1, 1),
        (1, 5, 1),
        (1, 1, 5),
        (9, 1, 1),
        (5, 5, 1),
        (5, 1, 5),
        (1, 9, 1),
        (1, 5, 5),
        (1, 1, 9),
    ]
    for scale_factors, circuit in zip(folding_pattern, scaled_circuits):
        scale_layer1, scale_layer2, scale_layer3 = scale_factors
        expected_circuit = Circuit(
            [ops.H.on_each(*qreg1)] * scale_layer1,
            [ops.CNOT.on(qreg1[0], qreg1[1]), ops.X.on(qreg1[2])]
            * scale_layer2,
            [ops.TOFFOLI.on(*qreg1)] * scale_layer3,
        )
        assert expected_circuit == circuit


@pytest.mark.parametrize(
    "test_input, expected",
    [(test_circuit1, 3), (test_circuit1_with_measurements, 3)],
)
def test_get_num_layers(test_input, expected):
    """Verifies function works as expected."""
    calculated_num_layers = _get_num_layers_without_measurements(test_input)

    assert calculated_num_layers == expected


@pytest.mark.parametrize(
    "test_input, test_chunks",
    [
        (test_circuit1 + test_circuit1 + test_circuit1, 3),
        (test_circuit1 + test_circuit1 + test_circuit1, 1),
        (test_circuit1 + test_circuit1 + test_circuit1, 5),
    ],
)
def test_get_num_chunks(test_input, test_chunks):
    """Verifies the chunking function works as expected."""
    assert test_chunks == len(_get_chunks(test_input, test_chunks))


def test_layers_with_chunking():
    """Checks the order of moments in the input circuit is unchanged with
    chunking."""

    test_circuit = test_circuit1 + test_circuit1 + test_circuit1
    calculated_circuit_chunks = _get_chunks(test_circuit, 4)
    expected_chunks = [
        test_circuit[0:3],
        test_circuit[3:5],
        test_circuit[5:7],
        test_circuit[7:],
    ]
    assert calculated_circuit_chunks == expected_chunks


@pytest.mark.parametrize(
    "test_input, degree, test_fold_multiplier, expected_scale_factor_vectors",
    [
        (test_circuit1, 1, 1, [(1, 1, 1), (3, 1, 1), (1, 3, 1), (1, 1, 3)]),
        (
            test_circuit1,
            2,
            1,
            [
                (1, 1, 1),
                (3, 1, 1),
                (1, 3, 1),
                (1, 1, 3),
                (5, 1, 1),
                (3, 3, 1),
                (3, 1, 3),
                (1, 5, 1),
                (1, 3, 3),
                (1, 1, 5),
            ],
        ),
        (
            test_circuit1,
            2,
            2,
            [
                (1, 1, 1),
                (5, 1, 1),
                (1, 5, 1),
                (1, 1, 5),
                (9, 1, 1),
                (5, 5, 1),
                (5, 1, 5),
                (1, 9, 1),
                (1, 5, 5),
                (1, 1, 9),
            ],
        ),
        (
            test_circuit1,
            2,
            3,
            [
                (1, 1, 1),
                (7, 1, 1),
                (1, 7, 1),
                (1, 1, 7),
                (13, 1, 1),
                (7, 7, 1),
                (7, 1, 7),
                (1, 13, 1),
                (1, 7, 7),
                (1, 1, 13),
            ],
        ),
        (
            test_circuit1_with_measurements,
            1,
            1,
            [(1, 1, 1), (3, 1, 1), (1, 3, 1), (1, 1, 3)],
        ),
        (
            test_circuit1_with_measurements,
            2,
            1,
            [
                (1, 1, 1),
                (3, 1, 1),
                (1, 3, 1),
                (1, 1, 3),
                (5, 1, 1),
                (3, 3, 1),
                (3, 1, 3),
                (1, 5, 1),
                (1, 3, 3),
                (1, 1, 5),
            ],
        ),
        (
            test_circuit1_with_measurements,
            2,
            2,
            [
                (1, 1, 1),
                (5, 1, 1),
                (1, 5, 1),
                (1, 1, 5),
                (9, 1, 1),
                (5, 5, 1),
                (5, 1, 5),
                (1, 9, 1),
                (1, 5, 5),
                (1, 1, 9),
            ],
        ),
        (
            test_circuit1_with_measurements,
            2,
            3,
            [
                (1, 1, 1),
                (7, 1, 1),
                (1, 7, 1),
                (1, 1, 7),
                (13, 1, 1),
                (7, 7, 1),
                (7, 1, 7),
                (1, 13, 1),
                (1, 7, 7),
                (1, 1, 13),
            ],
        ),
    ],
)
def test_get_scale_factor_vectors_no_chunking(
    test_input, degree, test_fold_multiplier, expected_scale_factor_vectors
):
    """Verifies vectors of scale factors are calculated accurately."""
    calculated_scale_factor_vectors = get_scale_factor_vectors(
        test_input, degree, test_fold_multiplier
    )

    assert calculated_scale_factor_vectors == expected_scale_factor_vectors


@pytest.mark.parametrize(
    "test_input, degree, test_fold_multiplier, test_chunks, expected_size",
    [
        (test_circuit1, 1, 1, 2, 3),
        (test_circuit1, 2, 1, 3, 10),
        (test_circuit1, 2, 3, 2, 6),
    ],
)
def test_get_scale_factor_vectors_with_chunking(
    test_input, degree, test_fold_multiplier, test_chunks, expected_size
):
    """Verifies vectors of scale factors are calculated accurately."""
    calculated_scale_factor_vectors = get_scale_factor_vectors(
        test_input, degree, test_fold_multiplier, test_chunks
    )

    assert len(calculated_scale_factor_vectors) == expected_size


@pytest.mark.parametrize(
    "test_input, num_chunks, error_msg",
    [
        (
            test_circuit1,
            0,
            "Number of chunks should be greater than or equal to 1.",
        ),
        (
            test_circuit1,
            5,
            "Number of chunks 5 cannot be greater than the number of layers"
            " 3.",
        ),
        (
            test_circuit1,
            -1,
            "Number of chunks should be greater than or equal to 1.",
        ),
    ],
)
def test_invalid_num_chunks(test_input, num_chunks, error_msg):
    """Ensures that the number of intended chunks in the input circuit raises
    an error for an invalid value."""
    with pytest.raises(ValueError, match=error_msg):
        get_scale_factor_vectors(test_input, 2, 2, num_chunks)


@pytest.mark.parametrize("frontend", SUPPORTED_PROGRAM_TYPES.keys())
def test_get_scale_factor_vectors_with_QPROGRAM(frontend):
    """Ensures the scale factor vectors are correctly generated for all
    supported frontends."""
    from mitiq.benchmarks import generate_ghz_circuit

    circuit = generate_ghz_circuit(3, frontend)

    scale_factor_vectors = get_scale_factor_vectors(circuit, 2, 2)
    expected_scale_factor_vectors = [
        (1, 1, 1),
        (5, 1, 1),
        (1, 5, 1),
        (1, 1, 5),
        (9, 1, 1),
        (5, 5, 1),
        (5, 1, 5),
        (1, 9, 1),
        (1, 5, 5),
        (1, 1, 9),
    ]
    assert scale_factor_vectors == expected_scale_factor_vectors


@pytest.mark.parametrize(
    "test_input, test_degree, test_fold_multiplier, error_msg",
    [
        (
            test_circuit1,
            0,
            1,
            "Multinomial degree must be greater than or equal to 1.",
        ),
        (
            test_circuit1,
            1,
            0,
            "Fold multiplier must be greater than or equal to 1.",
        ),
    ],
)
def test_invalid_degree_fold_multiplier(
    test_input, test_degree, test_fold_multiplier, error_msg
):
    """Ensures that the args for the main noise scaling function raise
    an error for an invalid value."""
    with pytest.raises(ValueError, match=error_msg):
        multivariate_layer_scaling(
            test_input, test_degree, test_fold_multiplier
        )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import math

import numpy as np

import mitiq
from mitiq.shadows.shadows_utils import (
    batch_calibration_data,
    create_string,
    fidelity,
    n_measurements_opts_expectation_bound,
    n_measurements_tomography_bound,
    valid_bitstrings,
)


def test_create_string():
    str_len = 5
    loc_list = [1, 3]
    assert create_string(str_len, loc_list) == "01010"


def test_valid_bitstrings():
    num_qubits = 5
    bitstrings_on_5_qubits = valid_bitstrings(num_qubits)
    assert len(bitstrings_on_5_qubits) == 2**num_qubits
    assert all(b == "0" or b == "1" for b in bitstrings_on_5_qubits.pop())

    num_qubits = 4
    max_hamming_weight = 2
    bitstrings_on_3_qubits_hamming_2 = valid_bitstrings(
        num_qubits, max_hamming_weight
    )
    assert len(bitstrings_on_3_qubits_hamming_2) == sum(
        math.comb(num_qubits, i) for i in range(max_hamming_weight + 1)
    )  # sum_{i == 0}^{max_hamming_weight} (num_qubits choose i)


def test_batch_calibration_data():
    data = (["010", "110", "000", "001"], ["XXY", "ZYY", "ZZZ", "XYZ"])
    num_batches = 2
    for bits, paulis in batch_calibration_data(data, num_batches):
        assert len(bits) == len(paulis) == num_batches


def test_n_measurements_tomography_bound():
    assert n_measurements_tomography_bound(0.5, 2) == 2176
    assert n_measurements_tomography_bound(1.0, 1) == 136
    assert n_measurements_tomography_bound(0.1, 3) == 217599


def test_n_measurements_opts_expectation_bound():
    observables = [
        mitiq.PauliString("X"),
        mitiq.PauliString("Y"),
        mitiq.PauliString("Z"),
    ]
    N, K = n_measurements_opts_expectation_bound(0.5, observables, 0.1)
    assert isinstance(N, int)
    assert isinstance(K, int)


def test_fidelity():
    state_vector = np.array([0.5, 0.5, 0.5, 0.5])
    rho = np.eye(4) / 4
    assert np.isclose(
        fidelity(state_vector, rho), 0.25
    ), f"Expected 0.25, got {fidelity(state_vector, rho)}"



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mitiq.pec.sampling functions."""

import cirq
import numpy as np
import pytest
from cirq import (
    Circuit,
    Gate,
    LineQubit,
    NamedQubit,
    depolarize,
    measure,
    measure_each,
    ops,
)
from pyquil import Program, gates
from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister

from mitiq.pec import (
    NoisyOperation,
    OperationRepresentation,
    represent_operation_with_global_depolarizing_noise,
    sample_circuit,
    sample_sequence,
)
from mitiq.pec.channels import _circuit_to_choi, _operation_to_choi
from mitiq.pec.representations import (
    represent_operation_with_local_depolarizing_noise,
    represent_operations_in_circuit_with_local_depolarizing_noise,
)
from mitiq.utils import _equal

xcirq = Circuit(cirq.X(cirq.LineQubit(0)))
zcirq = Circuit(cirq.Z(cirq.LineQubit(0)))
cnotcirq = Circuit(cirq.CNOT(cirq.LineQubit(0), cirq.LineQubit(1)))
czcirq = Circuit(cirq.CZ(cirq.LineQubit(0), cirq.LineQubit(1)))


def test_sample_sequence_cirq():
    circuit = Circuit(cirq.H(LineQubit(0)))

    circuit.append(measure(LineQubit(0)))

    rep = OperationRepresentation(
        ideal=circuit,
        noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
        coeffs=[0.5, -0.5],
    )

    for _ in range(5):
        seqs, signs, norm = sample_sequence(circuit, representations=[rep])
        assert isinstance(seqs[0], Circuit)
        assert signs[0] in {1, -1}
        assert norm == 1.0


def test_sample_sequence_qiskit():
    qreg = QuantumRegister(1)
    circuit = QuantumCircuit(qreg)
    _ = circuit.h(qreg)

    xcircuit = QuantumCircuit(qreg)
    _ = xcircuit.x(qreg)

    zcircuit = QuantumCircuit(qreg)
    _ = zcircuit.z(qreg)

    noisy_xop = NoisyOperation(xcircuit)
    noisy_zop = NoisyOperation(zcircuit)

    rep = OperationRepresentation(
        ideal=circuit,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, -0.5],
    )

    for _ in range(5):
        seqs, signs, norm = sample_sequence(circuit, representations=[rep])
        assert isinstance(seqs[0], QuantumCircuit)
        assert signs[0] in {1, -1}
        assert norm == 1.0


def test_sample_sequence_pyquil():
    circuit = Program(gates.H(0))

    noisy_xop = NoisyOperation(Program(gates.X(0)))
    noisy_zop = NoisyOperation(Program(gates.Z(0)))

    rep = OperationRepresentation(
        ideal=circuit,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, -0.5],
    )

    for _ in range(50):
        seqs, signs, norm = sample_sequence(circuit, representations=[rep])
        assert isinstance(seqs[0], Program)
        assert signs[0] in {1, -1}
        assert norm == 1.0


@pytest.mark.parametrize("seed", (1, 2, 3, 5))
def test_sample_sequence_cirq_random_state(seed):
    circuit = Circuit(cirq.H.on(LineQubit(0)))
    rep = OperationRepresentation(
        ideal=circuit,
        noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
        coeffs=[0.5, -0.5],
    )

    sequences, signs, norm = sample_sequence(
        circuit, [rep], random_state=np.random.RandomState(seed)
    )

    for _ in range(20):
        new_sequences, new_signs, new_norm = sample_sequence(
            circuit, [rep], random_state=np.random.RandomState(seed)
        )
        assert _equal(new_sequences[0], sequences[0])
        assert new_signs[0] == signs[0]
        assert np.isclose(new_norm, norm)


def test_qubit_independent_representation_cirq():
    """Test that an OperationRepresentation defined for some qubits can
    (optionally) be used to mitigate gates acting on different qubits."""
    circuit = Circuit([cirq.I.on(LineQubit(0)), cirq.H.on(LineQubit(1))])
    circuit.append(measure_each(*LineQubit.range(2)))

    rep = OperationRepresentation(
        ideal=Circuit(cirq.H.on(LineQubit(0))),
        noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
        coeffs=[0.5, -0.5],
        is_qubit_dependent=False,
    )

    expected_a = Circuit([cirq.I.on(LineQubit(0)), cirq.X.on(LineQubit(1))])
    expected_a.append(measure_each(*LineQubit.range(2)))

    expected_b = Circuit([cirq.I.on(LineQubit(0)), cirq.Z.on(LineQubit(1))])
    expected_b.append(measure_each(*LineQubit.range(2)))

    for _ in range(5):
        seqs, signs, norm = sample_circuit(circuit, representations=[rep])
        assert seqs[0] in [expected_a, expected_b]
        assert signs[0] in {1, -1}
        assert norm == 1.0


def test_qubit_independent_representation_qiskit():
    """Test that an OperationRepresentation defined for some qubits can
    (optionally) be used to mitigate gates acting on different qubits."""

    different_qreg = QuantumRegister(2, name="q")
    circuit_to_mitigate = QuantumCircuit(different_qreg)
    _ = circuit_to_mitigate.cx(*different_qreg)

    qreg = QuantumRegister(2, name="rep_register")
    xcircuit = QuantumCircuit(qreg)
    _ = xcircuit.x(qreg)

    zcircuit = QuantumCircuit(qreg)
    _ = zcircuit.z(qreg)

    noisy_xop = NoisyOperation(xcircuit)
    noisy_zop = NoisyOperation(zcircuit)

    ideal_op = QuantumCircuit(qreg)
    _ = ideal_op.cx(*qreg)
    rep = OperationRepresentation(
        ideal=ideal_op,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, -0.5],
        is_qubit_dependent=False,
    )

    # Expected outcomes
    xcircuit_different = QuantumCircuit(different_qreg)
    xcircuit_different.x(different_qreg)
    zcircuit_different = QuantumCircuit(different_qreg)
    zcircuit_different.z(different_qreg)

    for _ in range(5):
        seqs, signs, norm = sample_sequence(
            circuit_to_mitigate, representations=[rep]
        )
        assert seqs[0] in [xcircuit_different, zcircuit_different]
        assert signs[0] in {1, -1}
        assert norm == 1.0


def test_qubit_independent_representation_pyquil():
    """Test that an OperationRepresentation defined for some qubits can
    (optionally) be used to mitigate gates acting on different qubits."""
    circuit_to_mitigate = Program(gates.H(1))

    noisy_xop = NoisyOperation(Program(gates.X(0)))
    noisy_zop = NoisyOperation(Program(gates.Z(0)))

    rep = OperationRepresentation(
        ideal=Program(gates.H(0)),
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, -0.5],
        is_qubit_dependent=False,
    )

    for _ in range(50):
        seqs, signs, norm = sample_sequence(
            circuit_to_mitigate, representations=[rep]
        )
        assert seqs[0] in [Program(gates.X(1)), Program(gates.Z(1))]
        assert signs[0] in {1, -1}
        assert norm == 1.0


@pytest.mark.parametrize(
    "reps",
    [
        [],
        [
            OperationRepresentation(
                Circuit(cirq.H.on(LineQubit(0))),
                [NoisyOperation(xcirq), NoisyOperation(zcirq)],
                [0.5, -0.5],
            )
        ],
    ],
)
def test_sample_sequence_no_representation(reps):
    circuit = Circuit(cirq.H.on(LineQubit(0)), cirq.H.on(LineQubit(1)))
    circuit.append(measure_each(*LineQubit.range(2)))
    with pytest.warns(UserWarning, match="No representation found for"):
        sequences, signs, norm = sample_sequence(circuit, reps)
        assert _equal(sequences[0], circuit, require_qubit_equality=True)
        assert signs == [1]
        assert norm == 1


@pytest.mark.parametrize("measure", [True, False])
def test_sample_circuit_cirq(measure):
    circuit = Circuit(
        ops.H.on(LineQubit(0)),
        ops.CNOT.on(*LineQubit.range(2)),
    )
    if measure:
        circuit.append(measure_each(*LineQubit.range(2)))

    h_rep = OperationRepresentation(
        circuit[:1],
        [NoisyOperation(xcirq), NoisyOperation(zcirq)],
        [0.6, -0.6],
    )

    cnot_rep = OperationRepresentation(
        circuit[1:2],
        [NoisyOperation(cnotcirq), NoisyOperation(czcirq)],
        [0.7, -0.7],
    )

    for _ in range(50):
        sampled_circuits, signs, norm = sample_circuit(
            circuit, representations=[h_rep, cnot_rep]
        )

        assert isinstance(sampled_circuits[0], Circuit)
        assert signs[0] in (-1, 1)
        assert norm >= 1
        if measure:
            assert len(sampled_circuits[0]) == 3
            assert cirq.is_measurement(
                list(sampled_circuits[0].all_operations())[-1]  # last gate
            )
        else:
            assert len(sampled_circuits[0]) == 2


def test_sample_circuit_partial_representations():
    circuit = Circuit(
        ops.H.on(LineQubit(0)),
        ops.CNOT.on(*LineQubit.range(2)),
    )

    cnot_rep = OperationRepresentation(
        circuit[1:2],
        [NoisyOperation(cnotcirq), NoisyOperation(czcirq)],
        [0.7, -0.7],
    )

    for _ in range(10):
        with pytest.warns(UserWarning, match="No representation found for"):
            sampled_circuits, signs, norm = sample_circuit(
                circuit, representations=[cnot_rep]
            )
            assert isinstance(sampled_circuits[0], Circuit)
            assert len(sampled_circuits[0]) == 2
            assert signs[0] in (-1, 1)
            assert norm >= 1


def test_sample_circuit_pyquil():
    circuit = Program(gates.H(0), gates.CNOT(0, 1))

    h_rep = OperationRepresentation(
        circuit[:1],
        [
            NoisyOperation(Program(gates.X(0))),
            NoisyOperation(Program(gates.Z(0))),
        ],
        [0.6, -0.6],
    )

    cnot_rep = OperationRepresentation(
        circuit[1:],
        [
            NoisyOperation(Program(gates.CNOT(0, 1))),
            NoisyOperation(Program(gates.CZ(0, 1))),
        ],
        [0.7, -0.7],
    )

    for _ in range(50):
        sampled_circuits, signs, norm = sample_circuit(
            circuit, representations=[h_rep, cnot_rep]
        )

        assert isinstance(sampled_circuits[0], Program)
        assert len(sampled_circuits[0]) == 2
        assert signs[0] in (-1, 1)
        assert norm >= 1


def test_sample_circuit_with_seed():
    circ = Circuit([cirq.X.on(LineQubit(0)) for _ in range(10)])
    rep = OperationRepresentation(
        ideal=Circuit(cirq.X.on(LineQubit(0))),
        noisy_operations=[NoisyOperation(zcirq), NoisyOperation(xcirq)],
        coeffs=[1.0, -1.0],
    )

    expected_circuits, expected_signs, expected_norm = sample_circuit(
        circ, [rep], random_state=4
    )

    # Check we're not sampling the same operation every call to sample_sequence
    assert len(set(expected_circuits[0].all_operations())) > 1

    for _ in range(10):
        sampled_circuits, sampled_signs, sampled_norm = sample_circuit(
            circ, [rep], random_state=4
        )
        assert _equal(sampled_circuits[0], expected_circuits[0])
        assert sampled_signs[0] == expected_signs[0]
        assert sampled_norm == expected_norm


def test_sample_circuit_trivial_decomposition():
    circuit = Circuit(ops.H.on(NamedQubit("Q")))
    rep = OperationRepresentation(
        ideal=circuit,
        noisy_operations=[NoisyOperation(circuit)],
        coeffs=[1.0],
    )

    sampled_circuits, signs, norm = sample_circuit(
        circuit, [rep], random_state=1
    )
    assert _equal(sampled_circuits[0], circuit)
    assert signs[0] == 1
    assert np.isclose(norm, 1)


BASE_NOISE = 0.02
qreg = LineQubit.range(2)


@pytest.mark.parametrize("gate", [cirq.Y, cirq.CNOT])
def test_sample_sequence_choi(gate: Gate):
    """Tests the sample_sequence by comparing the exact Choi matrices."""
    qreg = LineQubit.range(gate.num_qubits())
    ideal_op = gate.on(*qreg)
    ideal_circ = Circuit(ideal_op)
    noisy_op_tree = [ideal_op] + [depolarize(BASE_NOISE)(q) for q in qreg]

    ideal_choi = _operation_to_choi(ideal_op)
    noisy_choi = _operation_to_choi(noisy_op_tree)

    representation = represent_operation_with_local_depolarizing_noise(
        ideal_circ,
        BASE_NOISE,
    )

    choi_unbiased_estimates = []
    rng = np.random.RandomState(1)
    for _ in range(500):
        imp_seqs, signs, norm = sample_sequence(
            ideal_circ, [representation], random_state=rng
        )
        noisy_sequence = imp_seqs[0].with_noise(depolarize(BASE_NOISE))
        sequence_choi = _circuit_to_choi(noisy_sequence)
        choi_unbiased_estimates.append(norm * signs[0] * sequence_choi)

    choi_pec_estimate = np.average(choi_unbiased_estimates, axis=0)
    noise_error = np.linalg.norm(ideal_choi - noisy_choi)
    pec_error = np.linalg.norm(ideal_choi - choi_pec_estimate)

    assert pec_error < noise_error
    assert np.allclose(ideal_choi, choi_pec_estimate, atol=0.05)


def test_sample_circuit_choi():
    """Tests the sample_circuit by comparing the exact Choi matrices."""
    # A simple 2-qubit circuit
    qreg = LineQubit.range(2)
    ideal_circ = Circuit(
        cirq.X.on(qreg[0]),
        cirq.I.on(qreg[1]),
        cirq.CNOT.on(*qreg),
    )

    noisy_circuit = ideal_circ.with_noise(depolarize(BASE_NOISE))

    ideal_choi = _circuit_to_choi(ideal_circ)
    noisy_choi = _operation_to_choi(noisy_circuit)

    rep_list = represent_operations_in_circuit_with_local_depolarizing_noise(
        ideal_circuit=ideal_circ,
        noise_level=BASE_NOISE,
    )

    choi_unbiased_estimates = []
    rng = np.random.RandomState(1)
    for _ in range(500):
        imp_circs, signs, norm = sample_circuit(
            ideal_circ, rep_list, random_state=rng
        )
        noisy_imp_circ = imp_circs[0].with_noise(depolarize(BASE_NOISE))
        sequence_choi = _circuit_to_choi(noisy_imp_circ)
        choi_unbiased_estimates.append(norm * signs[0] * sequence_choi)

    choi_pec_estimate = np.average(choi_unbiased_estimates, axis=0)
    noise_error = np.linalg.norm(ideal_choi - noisy_choi)
    pec_error = np.linalg.norm(ideal_choi - choi_pec_estimate)

    assert pec_error < noise_error
    assert np.allclose(ideal_choi, choi_pec_estimate, atol=0.05)


def test_conversions_in_sample_circuit():
    """Tests sample_circuit preserves idle qubits and quantum registers."""
    qreg = QuantumRegister(3, name="Q")
    creg = ClassicalRegister(2, name="C")
    circuit = QuantumCircuit(qreg, creg)
    circuit.h(qreg[0])
    circuit.cx(qreg[0], qreg[1])
    circuit.measure(0, 0)
    cnot_circuit = QuantumCircuit(qreg)
    cnot_circuit.cx(qreg[0], qreg[1])
    rep = represent_operation_with_global_depolarizing_noise(
        cnot_circuit,
        noise_level=0.0,
    )
    out_circuits, signs, norm = sample_circuit(circuit, [rep], num_samples=3)
    for out_circ in out_circuits:
        out_circ == circuit
    assert len(signs) == 3
    assert set(signs).issubset({1.0, -1.0})
    assert np.isclose(norm, 1.0)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest
from cirq import AmplitudeDampingChannel, Circuit, Gate, H, LineQubit, X, Y, Z

from mitiq.interface import convert_from_mitiq
from mitiq.pec.channels import _circuit_to_choi, _operation_to_choi
from mitiq.pec.representations.damping import (
    _represent_operation_with_amplitude_damping_noise,
    amplitude_damping_kraus,
)


@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", [X, Y, Z, H])
def test_single_qubit_representation_norm(gate: Gate, noise: float):
    q = LineQubit(0)
    optimal_norm = (1 + noise) / (1 - noise)
    norm = _represent_operation_with_amplitude_damping_noise(
        Circuit(gate(q)),
        noise,
    ).norm
    assert np.isclose(optimal_norm, norm)


# When _represent_operation_with_amplitude_damping_noise will
# support more circuit types we can add them below.
@pytest.mark.parametrize("circuit_type", ["cirq"])
@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", [X, Y, Z, H])
def test_amplitude_damping_representation_with_choi(
    gate: Gate,
    noise: float,
    circuit_type: str,
):
    """Tests the representation by comparing exact Choi matrices."""
    q = LineQubit(0)
    ideal_circuit = convert_from_mitiq(Circuit(gate.on(q)), circuit_type)
    ideal_choi = _circuit_to_choi(Circuit(gate.on(q)))
    op_rep = _represent_operation_with_amplitude_damping_noise(
        ideal_circuit,
        noise,
    )
    choi_components = []
    for coeff, noisy_op in op_rep.basis_expansion:
        implementable_circ = noisy_op.circuit
        depolarizing_op = AmplitudeDampingChannel(noise).on(q)
        # Apply noise after each sequence.
        # NOTE: noise is not applied after each operation.
        implementable_circ.append(depolarizing_op)
        sequence_choi = _operation_to_choi(implementable_circ)
        choi_components.append(coeff * sequence_choi)

    combination_choi = np.sum(choi_components, axis=0)
    assert np.allclose(ideal_choi, combination_choi, atol=10**-8)


def test_damping_kraus():
    expected = [[[1.0, 0.0], [0.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]]]
    assert np.allclose(amplitude_damping_kraus(1, 1), expected)
    expected = [
        [[1.0, 0.0], [0.0, np.sqrt(0.5)]],
        [[0.0, np.sqrt(0.5)], [0.0, 0.0]],
    ]
    assert np.allclose(amplitude_damping_kraus(0.5, 1), expected)
    # Test normalization of kraus operators
    for num_qubits in (1, 2, 3):
        for noise_level in (0.1, 1):
            kraus_ops = amplitude_damping_kraus(noise_level, num_qubits)
            dual_channel = sum([k.conj().T @ k for k in kraus_ops])
            assert np.allclose(dual_channel, np.eye(2**num_qubits))



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for Qibo <-> Cirq conversions."""

import cirq
import numpy as np
import pytest
import qibo
from qibo.models.circuit import Circuit as QiboCircuit

from mitiq.interface.mitiq_qibo import (
    UnsupportedQiboCircuitError,
    from_qibo,
    to_qibo,
)
from mitiq.utils import _equal


def test_from_qibo():
    qibo_circuit = QiboCircuit(2)
    qibo_circuit.add(qibo.gates.CNOT(0, 1))
    qibo_circuit.add(qibo.gates.M(0))

    circuit = from_qibo(qibo_circuit)

    correct = cirq.Circuit(cirq.CNOT(*cirq.LineQubit.range(2)))
    correct.append(cirq.measure(cirq.LineQubit(0)))

    assert _equal(circuit, correct, require_qubit_equality=False)


def test_from_qibo_register_name_error():
    qibo_circuit = QiboCircuit(2)
    qibo_circuit.add(qibo.gates.CNOT(0, 1))
    qibo_circuit.add(qibo.gates.M(0, register_name="K"))

    with pytest.raises(
        UnsupportedQiboCircuitError,
        match=(
            "OpenQASM does not support capital letters in "
            "register names but K was used."
        ),
    ):
        from_qibo(qibo_circuit)


def test_from_qibo_unsupported_multi_controlled_gate():
    qibo_circuit = QiboCircuit(4)
    qibo_circuit.add(qibo.gates.X(0).controlled_by(1, 2, 3))
    with pytest.raises(
        UnsupportedQiboCircuitError,
        match="OpenQASM does not support multi-controlled gates.",
    ):
        from_qibo(qibo_circuit)


def test_from_qibo_unsupported_gate():
    qibo_circuit = QiboCircuit(2)
    qibo_circuit.add(qibo.gates.fSim(0, 1, 0.4, 0.4))
    with pytest.raises(
        UnsupportedQiboCircuitError,
        match="fsim is not supported by OpenQASM.",
    ):
        from_qibo(qibo_circuit)


def test_from_qibo_unknown_cirq_gate():
    qibo_circuit = QiboCircuit(2)
    qibo_circuit.add(qibo.gates.CRY(0, 1, 0.4))
    qibo_circuit.add(qibo.gates.M(1))
    circuit = from_qibo(qibo_circuit)

    q0, q1 = cirq.LineQubit.range(2)
    correct = cirq.Circuit()
    correct.append(cirq.Ry(rads=0.2).on(q1))
    correct.append(cirq.CNOT(q0, q1))
    correct.append(cirq.Ry(rads=-0.2).on(q1))
    correct.append(cirq.CNOT(q0, q1))
    correct.append(cirq.measure(q1))

    assert _equal(circuit, correct, require_qubit_equality=False)


@pytest.mark.parametrize("random_state", range(10))
def test_to_from_qibo(random_state):
    circuit = cirq.testing.random_circuit(
        qubits=4, n_moments=2, op_density=1, random_state=random_state
    )

    converted = from_qibo(to_qibo(circuit))
    # Gates (e.g. iSWAP) aren't guaranteed to be preserved. Check unitary
    # instead of circuit equality.
    cirq.testing.assert_allclose_up_to_global_phase(
        cirq.unitary(converted), cirq.unitary(circuit), atol=1e-7
    )


def test_to_from_qibo_cnot_same_gates():
    qreg = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.CNOT(*qreg))
    converted = from_qibo(to_qibo(circuit))
    assert _equal(circuit, converted, require_qubit_equality=False)


def test_to_from_qibo_identity():
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.I(q))
    # Identity gate
    converted = from_qibo(to_qibo(circuit))
    assert _equal(circuit, converted, require_qubit_equality=False)


@pytest.mark.parametrize("i", range(10))
def test_qibo_integration(i):
    gates = [
        qibo.gates.X(0),
        qibo.gates.Y(0),
        qibo.gates.Z(0),
        qibo.gates.S(0),
        qibo.gates.SDG(0),
        qibo.gates.T(0),
        qibo.gates.I(0),
        qibo.gates.TDG(0),
        qibo.gates.RX(0, 0.4),
        qibo.gates.RY(0, 0.4),
        qibo.gates.RZ(0, 0.4),
        qibo.gates.H(0),
        qibo.gates.SX(0),
        qibo.gates.SXDG(0),
        qibo.gates.CSX(0, 1),
        qibo.gates.CSXDG(0, 1),
        qibo.gates.TOFFOLI(0, 1, 2),
        qibo.gates.SWAP(0, 1),
        qibo.gates.iSWAP(0, 1),
        qibo.gates.FSWAP(0, 1),
        qibo.gates.CNOT(0, 1),
        qibo.gates.CZ(0, 1),
        qibo.gates.U1(0, 0.4),
        qibo.gates.U2(0, 0.4, 0.5),
        qibo.gates.U3(0, 0.4, 0.5, 0.6),
        qibo.gates.CU1(0, 1, 0.4),
        qibo.gates.CU3(0, 1, 0.4, 0.5, 0.6),
        qibo.gates.CRX(0, 1, 0.4),
        qibo.gates.CRY(0, 1, 0.4),
        qibo.gates.CRZ(0, 1, 0.4),
        qibo.gates.RXX(0, 1, 0.4),
        qibo.gates.RYY(0, 1, 0.4),
        qibo.gates.RZZ(0, 1, 0.4),
    ]

    layers = 3
    np.random.seed(13 * i)
    gates_per_layers = [np.random.permutation(gates) for _ in range(layers)]

    qibo_circuit = QiboCircuit(3)
    for gates in gates_per_layers:
        for gate in gates:
            qibo_circuit.add(gate)

    base_circ = from_qibo(qibo_circuit)
    qibo_recovered = to_qibo(base_circ)
    circ_recovered = from_qibo(qibo_recovered)
    u_1 = cirq.unitary(base_circ)
    u_2 = cirq.unitary(circ_recovered)
    cirq.testing.assert_allclose_up_to_global_phase(u_1, u_2, atol=0)



"""Unit tests for the LRE extrapolation methods."""

import math
import random
import re
from typing import List
from unittest.mock import Mock

import numpy as np
import pytest
from cirq import DensityMatrixSimulator, depolarize

from mitiq import SUPPORTED_PROGRAM_TYPES, Executor, benchmarks
from mitiq.interface import mitiq_cirq
from mitiq.interface.mitiq_cirq import compute_density_matrix
from mitiq.lre import (
    combine_results,
    construct_circuits,
    execute_with_lre,
    lre_decorator,
)
from mitiq.lre.multivariate_scaling.layerwise_folding import (
    _get_chunks,
    multivariate_layer_scaling,
)
from mitiq.observable import Observable, PauliString
from mitiq.zne.scaling import fold_all, fold_global

# default circuit for all unit tests
test_cirq = benchmarks.generate_rb_circuits(
    n_qubits=1,
    num_cliffords=2,
)[0]

test_cirq_two_qubits = benchmarks.generate_rb_circuits(
    n_qubits=2,
    num_cliffords=2,
)[0]

obs_x = Observable(PauliString(spec="X"))
obs_y = Observable(PauliString(spec="Y"))
obs_z = Observable(PauliString(spec="Z"))
obs_zz = Observable(PauliString(spec="ZZ"))


# default execute function for all unit tests
def execute(circuit, noise_level=0.025) -> float:
    """Default executor for all unit tests."""
    noisy_circuit = circuit.with_noise(depolarize(p=noise_level))
    rho = DensityMatrixSimulator().simulate(noisy_circuit).final_density_matrix
    return rho[0, 0].real


def executor_density_matrix_typed(circuit) -> np.ndarray:
    return compute_density_matrix(circuit, noise_level=(0,))


def executor_density_matrix_batched(circuits) -> List[np.ndarray]:
    return [executor_density_matrix_typed(circuit) for circuit in circuits]


def batched_executor(circuits) -> List[float]:
    return [execute(circuit) for circuit in circuits]


noisy_val = execute(test_cirq)
ideal_val = execute(test_cirq, noise_level=0)


@pytest.mark.parametrize("degree, fold_multiplier", [(2, 2), (2, 3), (3, 4)])
def test_lre_exp_value(degree, fold_multiplier):
    """Verify LRE executors works as expected."""
    lre_exp_val = execute_with_lre(
        test_cirq,
        execute,
        degree=degree,
        fold_multiplier=fold_multiplier,
    )
    assert abs(lre_exp_val - ideal_val) <= abs(noisy_val - ideal_val)


@pytest.mark.parametrize(
    "circuit, degree, fold_multiplier, observable",
    [
        (test_cirq, 2, 3, obs_z),
        (test_cirq_two_qubits, 1, 2, obs_zz),
    ],
)
def test_lre_exp_value_with_observable(
    circuit, degree, fold_multiplier, observable
):
    """Verify LRE can be used with observables."""
    test_executor = Executor(mitiq_cirq.compute_density_matrix)
    lre_exp_val = execute_with_lre(
        circuit,
        test_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
        observable=observable,
    )

    assert isinstance(lre_exp_val, float)

    assert test_executor.calls_to_executor == len(
        multivariate_layer_scaling(circuit, degree, fold_multiplier)
    )


@pytest.mark.parametrize(
    "degree, fold_multiplier",
    [(2, 2), (2, 3), (3, 4)],
)
def test_lre_executor(degree, fold_multiplier):
    """Verify LRE batch executor works as expected."""
    test_executor = Executor(execute)
    lre_exp_val = execute_with_lre(
        test_cirq,
        test_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
    )
    assert isinstance(lre_exp_val, float)

    assert test_executor.calls_to_executor == len(
        multivariate_layer_scaling(test_cirq, degree, fold_multiplier)
    )


@pytest.mark.parametrize(
    "degree, fold_multiplier",
    [(2, 2), (2, 3), (3, 4)],
)
def test_lre_batched_executor(degree, fold_multiplier):
    """Verify LRE batch executor works as expected."""
    test_batched_executor = Executor(batched_executor, max_batch_size=200)
    lre_exp_val_batched = execute_with_lre(
        test_cirq,
        test_batched_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
    )
    assert isinstance(lre_exp_val_batched, float)

    assert test_batched_executor.calls_to_executor == 1
    assert (
        test_batched_executor.executed_circuits
        == multivariate_layer_scaling(test_cirq, degree, fold_multiplier)
    )


@pytest.mark.parametrize(
    "degree, fold_multiplier, observable",
    [(2, 2, obs_x), (2, 3, obs_y), (3, 4, obs_z)],
)
def test_lre_batched_executor_with_observable(
    degree, fold_multiplier, observable
):
    """Verify LRE batch executor with observable works as expected."""
    test_batched_executor = Executor(
        executor_density_matrix_batched, max_batch_size=200
    )
    lre_exp_val_batched = execute_with_lre(
        test_cirq,
        test_batched_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
        observable=observable,
    )
    assert isinstance(lre_exp_val_batched, float)

    assert test_batched_executor.calls_to_executor == 1
    assert (
        test_batched_executor.executed_circuits
        == multivariate_layer_scaling(test_cirq, degree, fold_multiplier)
    )


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_lre_all_qprogram(circuit_type):
    """Verify LRE works with all supported frontends."""
    degree, fold_multiplier = 2, 3
    circuit = benchmarks.generate_ghz_circuit(3, circuit_type)
    depth = 3  # not all circuit types have a simple way to compute depth

    mock_executor = Mock(side_effect=lambda _: random.random())

    lre_exp_val = execute_with_lre(
        circuit,
        mock_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
    )

    assert isinstance(lre_exp_val, float)
    assert mock_executor.call_count == math.comb(degree + depth, degree)


def test_lre_decorator():
    """Verify LRE decorators work as expected."""

    @lre_decorator(degree=2, fold_multiplier=2)
    def execute(circuit, noise_level=0.025):
        noisy_circuit = circuit.with_noise(depolarize(p=noise_level))
        rho = (
            DensityMatrixSimulator()
            .simulate(noisy_circuit)
            .final_density_matrix
        )
        return rho[0, 0].real

    assert abs(execute(test_cirq) - ideal_val) <= abs(noisy_val - ideal_val)


def test_lre_decorator_raised_error():
    """Verify an error is raised when the required parameters for the decorator
    are not specified."""
    with pytest.raises(TypeError, match=re.escape("lre_decorator() missing")):

        @lre_decorator()
        def execute(circuit, noise_level=0.025):
            noisy_circuit = circuit.with_noise(depolarize(p=noise_level))
            rho = (
                DensityMatrixSimulator()
                .simulate(noisy_circuit)
                .final_density_matrix
            )
            return rho[0, 0].real

        assert abs(execute(test_cirq) - ideal_val) <= abs(
            noisy_val - ideal_val
        )


@pytest.mark.parametrize("input_method", [(fold_global), (fold_all)])
def test_lre_executor_with_different_folding_methods(input_method):
    """Verify the executor works as expected for using non-default unitary
    folding methods."""
    lre_exp_val = execute_with_lre(
        test_cirq,
        execute,
        degree=2,
        fold_multiplier=2,
        folding_method=input_method,
    )
    assert abs(lre_exp_val - ideal_val) <= abs(noisy_val - ideal_val)


def test_lre_runs_correct_number_of_circuits_when_chunking():
    """Verify execute_with_lre works as expected when chunking is used.
    Note that this does not validate performance of chunking."""

    mock_executor = Mock(side_effect=lambda _: random.random())

    test_cirq = benchmarks.generate_rb_circuits(n_qubits=1, num_cliffords=12)[
        0
    ]

    degree, fold_multiplier, num_chunks = 2, 2, 10

    lre_exp_val_chunking = execute_with_lre(
        test_cirq,
        mock_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
        num_chunks=num_chunks,
    )

    chunked_circ = _get_chunks(test_cirq, num_chunks=num_chunks)
    assert isinstance(lre_exp_val_chunking, float)
    assert mock_executor.call_count == math.comb(
        degree + len(chunked_circ), degree
    )


def test_two_stage_lre():
    """Verify construct_circuits generates the appropriate number of circuits
    and combine_results returns the same results as calling execute_with_lre"""

    degree, fold_multiplier = 2, 2

    circuits = construct_circuits(test_cirq, degree, fold_multiplier)

    np.random.seed(42)

    def executor(circuit):
        return np.random.random()

    results = [executor(circuit) for circuit in circuits]

    final_result = combine_results(results, test_cirq, degree, fold_multiplier)

    np.random.seed(42)
    test_executor = Executor(executor)
    lre_exp_val = execute_with_lre(
        test_cirq,
        test_executor,
        degree=degree,
        fold_multiplier=fold_multiplier,
    )

    assert np.isclose(final_result, lre_exp_val)

    assert len(circuits) == len(test_executor.executed_circuits)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for DDD rules."""

import pytest
from cirq import CNOT, Circuit, I, LineQubit, X, Y, Z, bit_flip

from mitiq.ddd.rules.rules import general_rule, repeated_rule, xx, xyxy, yy
from mitiq.utils import _equal


@pytest.mark.parametrize(
    "slack_length",
    [4, 5, 8, 13, 21, 34],
)
@pytest.mark.parametrize(
    "gates",
    [
        [X, X],
        [X, Y, X, Y],
        [Y, Y],
        [X, Y, Z],
    ],
)
def test_general_sequences(slack_length, gates):
    sequence = general_rule(
        slack_length=slack_length,
        gates=gates,
    )
    gate_set = {X, Y, Z}
    seq_gates = [op.gate for op in sequence.all_operations()]
    assert len(sequence) == slack_length
    assert gates == [gate for gate in seq_gates if gate in gate_set]


@pytest.mark.parametrize(
    "slack_length",
    [5, 8, 13, 21, 34],
)
@pytest.mark.parametrize(
    "rule",
    [
        xx,
        xyxy,
        yy,
    ],
)
def test_built_in_sequences(rule, slack_length):
    name = rule.__name__
    sequence = rule(slack_length)
    gates = [X if i == "x" else Y for i in name]
    gate_set = {X, Y}
    seq_gates = [op.gate for op in sequence.all_operations()]
    assert len(sequence) == slack_length
    assert gates == [gate for gate in seq_gates if gate in gate_set]


@pytest.mark.parametrize(
    ("slack_length", "rule", "sequence"),
    [
        (
            5,
            xx,
            Circuit(
                [
                    X(LineQubit(0)) if i % 2 else I(LineQubit(0))
                    for i in range(5)
                ]
            ),
        ),
        (
            5,
            yy,
            Circuit(
                [
                    Y(LineQubit(0)) if i % 2 else I(LineQubit(0))
                    for i in range(5)
                ]
            ),
        ),
        (
            4,
            xyxy,
            Circuit(
                [
                    Y(LineQubit(0)) if i % 2 else X(LineQubit(0))
                    for i in range(4)
                ]
            ),
        ),
    ],
)
def test_exact_sequences(slack_length, rule, sequence):
    sequence_to_test = rule(slack_length)
    assert _equal(sequence_to_test, sequence)


@pytest.mark.parametrize(
    "slack_length",
    [1, 2, 3, 5, 8, 13, 21, 34],
)
@pytest.mark.parametrize("spacing", [i for i in range(5, 7)])
def test_rule_failures(slack_length, spacing):
    num_decoupling_gates = 3
    if slack_length < num_decoupling_gates:
        sequence = general_rule(
            slack_length=slack_length,
            spacing=spacing,
            gates=[X, Y, Z],
        )
        assert len(sequence) == 0
    elif slack_length < (
        (num_decoupling_gates + 1) * spacing + num_decoupling_gates
    ):
        sequence = general_rule(
            slack_length=slack_length,
            spacing=spacing,
            gates=[X, Y, Z],
        )
        assert len(sequence) == 0
    else:
        sequence = general_rule(
            slack_length=slack_length,
            spacing=spacing,
            gates=[X, Y, Z],
        )
        assert len(sequence) == slack_length


@pytest.mark.parametrize(
    "slack_length",
    [1, 2, 3, 5],
)
@pytest.mark.parametrize(
    "gates",
    [
        [X],
        [Y],
        [Z],
    ],
)
def test_general_for_incomplete_rule(slack_length, gates):
    with pytest.raises(ValueError, match="too short to make a ddd sequence"):
        general_rule(
            slack_length=slack_length,
            gates=gates,
        )


@pytest.mark.parametrize(
    "slack_length",
    [3, 5],
)
@pytest.mark.parametrize(
    "gates",
    [
        [CNOT, X, Y],
    ],
)
def test_general_for_multi_qubit_gate(slack_length, gates):
    with pytest.raises(ValueError, match="Wrong number of qubits"):
        general_rule(
            slack_length=slack_length,
            gates=gates,
        )


@pytest.mark.parametrize(
    "slack_length",
    [4, 5, 8, 13, 21, 34],
)
@pytest.mark.parametrize(
    "gates",
    [
        [X, X],
        [X, Y, X, Y],
        [Y, Y],
        [X, Y, Z],
    ],
)
def test_repeated_sequences(slack_length, gates):
    sequence = repeated_rule(
        slack_length=slack_length,
        gates=gates,
    )
    num_reps = slack_length // len(gates)
    gate_set = {X, Y, Z}
    seq_gates = [op.gate for op in sequence.all_operations()]
    assert len(sequence) == slack_length
    assert gates * num_reps == [gate for gate in seq_gates if gate in gate_set]


@pytest.mark.parametrize(
    "slack_length",
    [2, 3],
)
@pytest.mark.parametrize(
    "gates",
    [
        [X, Y, X, Y],
        [Y, Y, Y, Y],
    ],
)
def test_short_repeated_sequences(slack_length, gates):
    sequence = repeated_rule(
        slack_length=slack_length,
        gates=gates,
    )
    assert len(sequence) == 0


@pytest.mark.parametrize(
    "gates",
    [
        [bit_flip(p=0.1), bit_flip(p=0.1)],
        [X, X, X],
    ],
)
def test_not_unitary(gates):
    if bit_flip(p=0.1) in gates:
        with pytest.raises(TypeError, match="cirq.unitary failed"):
            general_rule(slack_length=17, gates=gates)
    else:
        with pytest.raises(
            ValueError, match="is not equivalent to the identity"
        ):
            general_rule(slack_length=17, gates=gates)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for randomized benchmarking circuits."""

import numpy as np
import pytest

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks.randomized_benchmarking import generate_rb_circuits


@pytest.mark.parametrize("n_qubits", (1, 2))
def test_rb_circuits(n_qubits):
    depth = 10

    # test single qubit RB
    for trials in [2, 3]:
        circuits = generate_rb_circuits(
            n_qubits=n_qubits, num_cliffords=depth, trials=trials
        )
        for qc in circuits:
            # we check the ground state population to ignore any global phase
            wvf = qc.final_state_vector()
            zero_prob = abs(wvf[0] ** 2)
            assert np.isclose(zero_prob, 1)


@pytest.mark.parametrize("n_qubits", (1, 2))
@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_rb_conversion(n_qubits, return_type):
    depth = 10
    for trials in [2, 3]:
        circuits = generate_rb_circuits(
            n_qubits=n_qubits,
            num_cliffords=depth,
            trials=trials,
            return_type=return_type,
        )
        for qc in circuits:
            assert return_type in qc.__module__


def test_bad_n_qubits():
    with pytest.raises(
        ValueError, match="Only generates RB circuits on one or two "
    ):
        for trials in [2, 3]:
            generate_rb_circuits(n_qubits=4, num_cliffords=10, trials=trials)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for readout confusion inversion."""

from functools import partial
from typing import List

import cirq
import numpy as np
import pytest
from cirq.experiments.single_qubit_readout_calibration_test import (
    NoisySingleQubitReadoutSampler,
)

from mitiq import Executor, MeasurementResult, Observable, PauliString
from mitiq.interface.mitiq_cirq import sample_bitstrings
from mitiq.raw import execute as raw_execute
from mitiq.rem import execute_with_rem, mitigate_executor, rem_decorator
from mitiq.rem.inverse_confusion_matrix import (
    generate_inverse_confusion_matrix,
)

# Default qubit register and circuit for unit tests
qreg = [cirq.LineQubit(i) for i in range(2)]
circ_with_measurements = cirq.Circuit(
    cirq.ops.X.on_each(*qreg), cirq.measure_each(*qreg)
)
circ = cirq.Circuit(cirq.ops.X.on_each(*qreg))
observable = Observable(PauliString("ZI"), PauliString("IZ"))


def noisy_readout_executor(
    circuit, p0: float = 0.01, p1: float = 0.01, shots: int = 8192
) -> MeasurementResult:
    simulator = NoisySingleQubitReadoutSampler(p0, p1)
    result = simulator.run(circuit, repetitions=shots)

    return MeasurementResult(
        result=np.column_stack(list(result.measurements.values())),
        qubit_indices=tuple(
            # q[2:-1] is necessary to convert "q(number)" into "number"
            int(q[2:-1])
            for k in result.measurements.keys()
            for q in k.split(",")
        ),
    )


npX = np.array([[0, 1], [1, 0]])
"""Defines the sigma_x Pauli matrix in SU(2) algebra as a (2,2) `np.array`."""

npZ = np.array([[1, 0], [0, -1]])
"""Defines the sigma_z Pauli matrix in SU(2) algebra as a (2,2) `np.array`."""


# An invalid executor for unit tests since it returns an expectation
def invalid_executor(circuit) -> float:
    wavefunction = circuit.final_state_vector(
        ignore_terminal_measurements=True
    )
    return np.real(wavefunction.conj().T @ np.kron(npX, npZ) @ wavefunction)


def test_rem_invalid_executor():
    identity = np.identity(4)
    with pytest.raises(TypeError, match="not of type MeasurementResult"):
        execute_with_rem(
            circ,
            invalid_executor,
            observable=None,
            inverse_confusion_matrix=identity,
        )


def test_rem_identity():
    executor = partial(sample_bitstrings, noise_level=(0,))
    identity = np.identity(4)
    result = execute_with_rem(
        circ, executor, observable, inverse_confusion_matrix=identity
    )
    assert np.isclose(result, -2.0)


def test_rem_with_matrix():
    # test with an executor that completely flips results
    p0 = 1
    p1 = 1
    noisy_executor = partial(noisy_readout_executor, p0=p0, p1=p1)
    unmitigated = raw_execute(circ, noisy_executor, observable)
    assert np.isclose(unmitigated, 2.0)

    inverse_confusion_matrix = np.array(
        [
            [0, 0, 0, 1],
            [0, 0, 1, 0],
            [0, 1, 0, 0],
            [1, 0, 0, 0],
        ]
    )

    mitigated = execute_with_rem(
        circ,
        Executor(noisy_executor),
        observable,
        inverse_confusion_matrix=inverse_confusion_matrix,
    )
    assert np.isclose(mitigated, -2.0)


def test_rem_with_invalid_matrix():
    executor = partial(sample_bitstrings, noise_level=(0,))
    identity = np.identity(2)
    with pytest.raises(ValueError):
        execute_with_rem(
            circ, executor, observable, inverse_confusion_matrix=identity
        )


def test_doc_is_preserved():
    """Tests that the doc of the original executor is preserved."""

    def first_executor(circuit):
        """Doc of the original executor."""
        return MeasurementResult()

    identity = np.identity(4)

    mit_executor = mitigate_executor(
        first_executor, inverse_confusion_matrix=identity
    )
    assert mit_executor.__doc__ == first_executor.__doc__

    @rem_decorator(inverse_confusion_matrix=identity)
    def second_executor(circuit):
        """Doc of the original executor."""
        return 0

    assert second_executor.__doc__ == first_executor.__doc__


@pytest.mark.parametrize(
    "p0, p1, atol",
    [
        (0, 0, 0),
        (1, 1, 0),
        (
            1,
            0,
            2.1,
        ),  # catastrophic failure from all 0s state, which is why atol=2.1
        (
            0,
            1,
            0,
        ),  # undetectable: results in all 1s state, which is the desired state
        (0.3, 0.1, 0),
        (0.1, 0.3, 0),
        (0.001, 0.001, 1e-3),
        (0.02, 0.04, 0),
        (0.3, 0.7, 0),
    ],
)
def test_mitigate_executor(p0, p1, atol):
    true_rem_value = -2.0

    noisy_executor = partial(noisy_readout_executor, p0=p0, p1=p1)

    num_qubits = len(qreg)
    inverse_confusion_matrix = generate_inverse_confusion_matrix(
        num_qubits, p0=p0, p1=p1
    )

    base = raw_execute(circ, noisy_executor, observable)

    mitigated_executor = mitigate_executor(
        noisy_executor,
        inverse_confusion_matrix=inverse_confusion_matrix,
    )
    rem_value = raw_execute(circ, mitigated_executor, observable)
    assert abs(true_rem_value - rem_value) <= abs(
        true_rem_value - base
    ) or np.isclose(
        abs(true_rem_value - rem_value),
        abs(true_rem_value - base),
        atol=atol,
    )


def test_rem_decorator():
    true_rem_value = -2.0

    num_qubits = len(qreg)

    # test with an executor that completely flips results
    p0 = 1
    p1 = 1
    inverse_confusion_matrix = generate_inverse_confusion_matrix(
        num_qubits, p0=p0, p1=p1
    )

    @rem_decorator(inverse_confusion_matrix=inverse_confusion_matrix)
    def noisy_readout_decorated_executor(qp) -> MeasurementResult:
        return noisy_readout_executor(qp, p0=p0, p1=p1)

    noisy_executor = partial(noisy_readout_executor, p0=p0, p1=p1)

    base = raw_execute(circ, noisy_executor, observable)

    rem_value = raw_execute(circ, noisy_readout_decorated_executor, observable)
    assert abs(true_rem_value - rem_value) < abs(true_rem_value - base)


def test_rem_decorator_batched():
    circuits = [circ, 3 * circ]
    true_values = [-2.0, -2.0]
    num_qubits = len(qreg)
    p0 = 0.2
    p1 = 0.1
    inverse_confusion_matrix = generate_inverse_confusion_matrix(
        num_qubits, p0=p0, p1=p1
    )

    @rem_decorator(inverse_confusion_matrix=inverse_confusion_matrix)
    def noisy_readout_batched(circuits) -> List[MeasurementResult]:
        return [noisy_readout_executor(c, p0=p0, p1=p1) for c in circuits]

    noisy_executor = partial(noisy_readout_executor, p0=p0, p1=p1)
    base_values = [
        raw_execute(c, noisy_executor, observable) for c in circuits
    ]
    rem_values = Executor(noisy_readout_batched).evaluate(circuits, observable)
    for true_val, base, rem_val in zip(true_values, base_values, rem_values):
        assert abs(true_val - rem_val) < abs(true_val - base)
        assert np.isclose(true_val, rem_val, atol=0.05)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for inverse confusion matrix helper functions."""

from functools import reduce
from math import isclose

import numpy as np
import pytest

from mitiq import MeasurementResult
from mitiq.rem.inverse_confusion_matrix import (
    bitstrings_to_probability_vector,
    closest_positive_distribution,
    generate_inverse_confusion_matrix,
    generate_tensored_inverse_confusion_matrix,
    mitigate_measurements,
    sample_probability_vector,
)


def test_sample_probability_vector_invalid_size():
    with pytest.raises(ValueError, match="power of 2"):
        sample_probability_vector([1 / 3, 1 / 3, 1 / 3], 3)


def test_sample_probability_vector_single_qubit():
    bitstrings = sample_probability_vector(np.array([1, 0]), 10)
    assert all(b == "0" for b in bitstrings)

    bitstrings = sample_probability_vector(np.array([0, 1]), 10)
    assert all(b == "1" for b in bitstrings)

    np.random.seed(0)
    bitstrings = sample_probability_vector(np.array([0.5, 0.5]), 1000)
    assert sum(int(b) for b in bitstrings) == 483


def test_sample_probability_vector_two_qubits():
    bitstrings = sample_probability_vector(np.array([1, 0, 0, 0]), 10)
    assert all(b == "00" for b in bitstrings)

    bitstrings = sample_probability_vector(np.array([0, 1, 0, 0]), 10)
    assert all(b == "01" for b in bitstrings)

    bitstrings = sample_probability_vector(np.array([0, 0, 1, 0]), 10)
    assert all(b == "10" for b in bitstrings)

    bitstrings = sample_probability_vector(np.array([0, 0, 0, 1]), 10)
    assert all(b == "11" for b in bitstrings)


def test_bitstrings_to_probability_vector():
    pv = bitstrings_to_probability_vector([[0]])
    assert (pv == np.array([1, 0])).all()

    pv = bitstrings_to_probability_vector([[1]])
    assert (pv == np.array([0, 1])).all()

    pv = bitstrings_to_probability_vector([[0], [1]])
    assert (pv == np.array([0.5, 0.5])).all()

    pv = bitstrings_to_probability_vector([[0, 0]])
    assert (pv == np.array([1, 0, 0, 0])).all()

    pv = bitstrings_to_probability_vector([[1, 1]])
    assert (pv == np.array([0, 0, 0, 1])).all()


@pytest.mark.parametrize("_", range(10))
def test_probability_vector_roundtrip(_):
    pv = np.random.rand(4)
    pv /= np.sum(pv)
    assert isclose(
        np.linalg.norm(
            pv
            - bitstrings_to_probability_vector(
                sample_probability_vector(pv, 1000)
            )
        ),
        0,
        abs_tol=0.1,
    )


def test_generate_inverse_confusion_matrix():
    num_qubits = 2
    identity = np.identity(4)
    assert (
        generate_inverse_confusion_matrix(num_qubits, p0=0, p1=0) == identity
    ).all()
    assert (
        generate_inverse_confusion_matrix(num_qubits, p0=1, p1=1)
        == np.flipud(identity)
    ).all()


@pytest.mark.parametrize(
    "num_qubits, confusion_matrices, expected",
    [
        (2, [np.identity(2), np.identity(2)], np.identity(4)),
        (2, [np.identity(4)], np.identity(4)),
        (3, [np.identity(4), np.identity(2)], np.identity(8)),
        # all are faulty qubits, flipping to opposite value
        (
            2,
            [np.flipud(np.identity(2)), np.flipud(np.identity(2))],
            np.flipud(np.identity(4)),
        ),
        (
            3,
            [np.flipud(np.identity(4)), np.flipud(np.identity(2))],
            np.flipud(np.identity(8)),
        ),
        # wrongly sized confusion matrices
        (3, [np.identity(2), np.identity(2)], ValueError),
        # one qubit flips values, one is perfect, one is random
        (
            3,
            [np.flipud(np.identity(2)), np.identity(2), 0.5 * np.ones((2, 2))],
            np.linalg.pinv(
                reduce(
                    np.kron,
                    [
                        np.flipud(np.identity(2)),
                        np.identity(2),
                        0.5 * np.ones((2, 2)),
                    ],
                )
            ),
        ),
    ],
)
def test_generate_tensored_inverse_confusion_matrix(
    num_qubits, confusion_matrices, expected
):
    if expected is ValueError:
        with pytest.raises(ValueError):
            generate_tensored_inverse_confusion_matrix(
                num_qubits, confusion_matrices
            )
    else:
        assert np.allclose(
            generate_tensored_inverse_confusion_matrix(
                num_qubits, confusion_matrices
            ),
            expected,
        )


def test_mitigate_measurements():
    identity = np.identity(4)

    measurements = MeasurementResult([[1, 0]])
    assert mitigate_measurements(measurements, identity) == measurements
    assert mitigate_measurements(measurements, np.flipud(identity)).result == [
        [0, 1]
    ]

    measurements = MeasurementResult([[0, 1]])
    assert mitigate_measurements(measurements, identity) == measurements
    assert mitigate_measurements(measurements, np.flipud(identity)).result == [
        [1, 0]
    ]


def test_closest_positive_distribution():
    inputs = [
        [0.3, 0.7],  # Test optimal input
        [-0.1, 1.1],  # Test negative elements
        [10, 10],  # Test normalization
        [-1, 1, -1, 1],  # Test more elements
        [-1, 0.1, -1, 0.2],  # Non-trivial problem
    ]
    expected = [
        [0.3, 0.7],
        [0, 1],
        [0.5, 0.5],
        [0, 0.5, 0, 0.5],
        [0, 0.450317, 0, 0.549683],
    ]
    for quasi_prob, prob in zip(inputs, expected):
        assert np.allclose(
            closest_positive_distribution(quasi_prob),
            prob,
            atol=1e-5,
        )


@pytest.mark.filterwarnings("ignore:invalid value encountered in divide")
def test_closest_positive_distribution_error():
    """Test unfeasible problem to trigger error."""
    with pytest.raises(ValueError, match="REM failed to determine"):
        closest_positive_distribution([0, 0])



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import cirq_ionq.ionq_native_gates as cirq_ionq_ops
import numpy as np
import pytest
from braket.circuits import Circuit as BKCircuit
from braket.circuits import Instruction
from braket.circuits import gates as braket_gates
from cirq import Circuit, LineQubit, ops, protocols, testing

from mitiq.interface.mitiq_braket.conversions import from_braket, to_braket
from mitiq.utils import _equal


@pytest.mark.parametrize(
    "qreg", (LineQubit.range(2), [LineQubit(1), LineQubit(6)])
)
def test_to_from_braket_bell_circuit(qreg):
    cirq_circuit = Circuit(ops.H(qreg[0]), ops.CNOT(*qreg))
    test_circuit = from_braket(to_braket(cirq_circuit))
    assert _equal(test_circuit, cirq_circuit, require_qubit_equality=True)


def test_from_braket_non_parameterized_single_qubit_gates():
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.I(), target=0),
        Instruction(braket_gates.X(), target=1),
        Instruction(braket_gates.Y(), target=2),
        Instruction(braket_gates.Z(), target=3),
        Instruction(braket_gates.H(), target=0),
        Instruction(braket_gates.S(), target=1),
        Instruction(braket_gates.Si(), target=2),
        Instruction(braket_gates.T(), target=3),
        Instruction(braket_gates.Ti(), target=0),
        Instruction(braket_gates.V(), target=1),
        Instruction(braket_gates.Vi(), target=2),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = from_braket(braket_circuit)

    for i, op in enumerate(cirq_circuit.all_operations()):
        assert np.allclose(
            instructions[i].operator.to_matrix(), protocols.unitary(op)
        )

    qreg = LineQubit.range(4)
    expected_cirq_circuit = Circuit(
        ops.I(qreg[0]),
        ops.X(qreg[1]),
        ops.Y(qreg[2]),
        ops.Z(qreg[3]),
        ops.H(qreg[0]),
        ops.S(qreg[1]),
        ops.S(qreg[2]) ** -1,
        ops.T(qreg[3]),
        ops.T(qreg[0]) ** -1,
        ops.X(qreg[1]) ** 0.5,
        ops.X(qreg[2]) ** -0.5,
    )
    assert _equal(cirq_circuit, expected_cirq_circuit)


@pytest.mark.parametrize("qubit_index", (0, 3))
def test_from_braket_parameterized_single_qubit_gates(qubit_index):
    braket_circuit = BKCircuit()
    pgates = [
        braket_gates.Rx,
        braket_gates.Ry,
        braket_gates.Rz,
        braket_gates.PhaseShift,
        braket_gates.GPi,
        braket_gates.GPi2,
    ]
    angles = np.random.RandomState(11).random(len(pgates))
    instructions = [
        Instruction(rot(a), target=qubit_index)
        for rot, a in zip(pgates, angles)
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = from_braket(braket_circuit)

    for i, op in enumerate(cirq_circuit.all_operations()):
        assert np.allclose(
            instructions[i].operator.to_matrix(), protocols.unitary(op)
        )

    qubit = LineQubit(qubit_index)
    expected_cirq_circuit = Circuit(
        ops.rx(angles[0]).on(qubit),
        ops.ry(angles[1]).on(qubit),
        ops.rz(angles[2]).on(qubit),
        ops.Z.on(qubit) ** (angles[3] / np.pi),
        cirq_ionq_ops.GPIGate(phi=angles[4] / (2 * np.pi)).on(qubit),
        cirq_ionq_ops.GPI2Gate(phi=angles[5] / (2 * np.pi)).on(qubit),
    )
    assert _equal(
        cirq_circuit, expected_cirq_circuit, require_qubit_equality=True
    )


def test_from_braket_non_parameterized_two_qubit_gates():
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.CNot(), target=[2, 3]),
        Instruction(braket_gates.Swap(), target=[3, 4]),
        Instruction(braket_gates.ISwap(), target=[2, 3]),
        Instruction(braket_gates.CZ(), target=(3, 4)),
        Instruction(braket_gates.CY(), target=(2, 3)),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = from_braket(braket_circuit)

    qreg = LineQubit.range(2, 5)
    expected_cirq_circuit = Circuit(
        ops.CNOT(*qreg[:2]),
        ops.SWAP(*qreg[1:]),
        ops.ISWAP(*qreg[:2]),
        ops.CZ(*qreg[1:]),
        ops.ControlledGate(ops.Y).on(*qreg[:2]),
    )
    assert np.allclose(
        protocols.unitary(cirq_circuit),
        protocols.unitary(expected_cirq_circuit),
    )


def test_from_braket_parameterized_two_qubit_gates():
    pgates = [
        braket_gates.CPhaseShift,
        braket_gates.CPhaseShift00,
        braket_gates.CPhaseShift01,
        braket_gates.CPhaseShift10,
        braket_gates.PSwap,
        braket_gates.XX,
        braket_gates.YY,
        braket_gates.ZZ,
        braket_gates.XY,
    ]
    angles = np.random.RandomState(2).random(len(pgates))
    instructions = [
        Instruction(rot(a), target=[0, 1]) for rot, a in zip(pgates, angles)
    ]

    cirq_circuits = list()
    for instr in instructions:
        braket_circuit = BKCircuit()
        braket_circuit.add_instruction(instr)
        cirq_circuits.append(from_braket(braket_circuit))

    for instr, cirq_circuit in zip(instructions, cirq_circuits):
        assert np.allclose(instr.operator.to_matrix(), cirq_circuit.unitary())


def test_from_braket_parameterized_two_qubit_two_parameters_gates():
    pgates = [
        braket_gates.MS,
    ]
    angles = np.random.RandomState(2).random((len(pgates), 2))
    instructions = [
        Instruction(rot(a[0], a[1]), target=[0, 1])
        for rot, a in zip(pgates, angles)
    ]

    cirq_circuits = list()
    for instr in instructions:
        braket_circuit = BKCircuit()
        braket_circuit.add_instruction(instr)
        cirq_circuits.append(from_braket(braket_circuit))

    for instr, cirq_circuit in zip(instructions, cirq_circuits):
        assert np.allclose(instr.operator.to_matrix(), cirq_circuit.unitary())


def test_from_braket_three_qubit_gates():
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.CCNot(), target=[1, 2, 3]),
        Instruction(braket_gates.CSwap(), target=[1, 2, 3]),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = from_braket(braket_circuit)

    qreg = LineQubit.range(1, 4)
    expected_cirq_circuit = Circuit(ops.TOFFOLI(*qreg), ops.FREDKIN(*qreg))
    assert np.allclose(
        protocols.unitary(cirq_circuit),
        protocols.unitary(expected_cirq_circuit),
    )


def _rotation_of_pi_over_7(num_qubits):
    matrix = np.identity(2**num_qubits)
    matrix[0:2, 0:2] = [
        [np.cos(np.pi / 7), np.sin(np.pi / 7)],
        [-np.sin(np.pi / 7), np.cos(np.pi / 7)],
    ]
    return matrix


def test_from_braket_raises_on_unsupported_gates():
    for num_qubits in range(1, 5):
        braket_circuit = BKCircuit()
        instr = Instruction(
            braket_gates.Unitary(_rotation_of_pi_over_7(num_qubits)),
            target=list(range(num_qubits)),
        )
        braket_circuit.add_instruction(instr)
        with pytest.raises(ValueError):
            from_braket(braket_circuit)


def test_to_braket_raises_on_unsupported_gates():
    for num_qubits in range(3, 5):
        qubits = [LineQubit(int(qubit)) for qubit in range(num_qubits)]
        op = ops.MatrixGate(_rotation_of_pi_over_7(num_qubits)).on(*qubits)
        circuit = Circuit(op)
        with pytest.raises(ValueError):
            to_braket(circuit)


def test_to_from_braket_common_one_qubit_gates():
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket -> Cirq.
    """
    rots = [ops.rx, ops.ry, ops.rz]
    angles = [1 / 5, 3 / 5, -4 / 5]
    qubit = LineQubit(0)
    cirq_circuit = Circuit(
        # Paulis.
        ops.X(qubit),
        ops.Y(qubit),
        ops.Z(qubit),
        # Single-qubit rotations.
        [rot(angle).on(qubit) for rot, angle in zip(rots, angles)],
        # Rz alter egos.
        ops.T(qubit),
        ops.T(qubit) ** -1,
        ops.S(qubit),
        ops.S(qubit) ** -1,
        # Rx alter egos.
        ops.X(qubit) ** 0.5,
        ops.X(qubit) ** -0.5,
    )

    test_circuit = from_braket(to_braket(cirq_circuit))
    assert _equal(test_circuit, cirq_circuit, require_qubit_equality=True)


@pytest.mark.parametrize(
    "uncommon_gate",
    [
        ops.HPowGate(exponent=-1 / 14),
        cirq_ionq_ops.GPIGate(phi=1 / 14),
        cirq_ionq_ops.GPI2Gate(phi=1 / 14),
    ],
)
def test_to_from_braket_uncommon_one_qubit_gates(uncommon_gate):
    """These gates get decomposed when converting Cirq -> Braket -> Cirq, but
    the unitaries should be equal up to global phase.
    """
    cirq_circuit = Circuit(uncommon_gate.on(LineQubit(0)))
    test_circuit = from_braket(to_braket(cirq_circuit))
    testing.assert_allclose_up_to_global_phase(
        protocols.unitary(test_circuit),
        protocols.unitary(cirq_circuit),
        atol=1e-7,
    )


@pytest.mark.parametrize(
    "common_gate",
    [
        ops.CNOT,
        ops.CZ,
        ops.ISWAP,
        ops.XXPowGate(exponent=-0.2),
        ops.YYPowGate(exponent=0.3),
        ops.ZZPowGate(exponent=-0.1),
    ],
)
def test_to_from_braket_common_two_qubit_gates(common_gate):
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket -> Cirq.
    """
    cirq_circuit = Circuit(common_gate.on(*LineQubit.range(2)))
    test_circuit = from_braket(to_braket(cirq_circuit))
    testing.assert_allclose_up_to_global_phase(
        protocols.unitary(test_circuit),
        protocols.unitary(cirq_circuit),
        atol=1e-7,
    )

    # Cirq AAPowGate has a different global phase than braket AA which gets
    # lost in translation. Here, AA = XX, YY, or ZZ.
    if not isinstance(
        common_gate, (ops.XXPowGate, ops.YYPowGate, ops.ZZPowGate)
    ):
        assert _equal(test_circuit, cirq_circuit, require_qubit_equality=True)


@pytest.mark.parametrize(
    "uncommon_gate",
    [
        ops.CNotPowGate(exponent=-1 / 17),
        ops.CZPowGate(exponent=2 / 7),
        cirq_ionq_ops.MSGate(phi0=1 / 7, phi1=2 / 7),
    ],
)
def test_to_from_braket_uncommon_two_qubit_gates(uncommon_gate):
    """These gates get decomposed when converting Cirq -> Braket -> Cirq, but
    the unitaries should be equal up to global phase.
    """
    cirq_circuit = Circuit(uncommon_gate.on(*LineQubit.range(2)))
    test_circuit = from_braket(to_braket(cirq_circuit))

    testing.assert_allclose_up_to_global_phase(
        protocols.unitary(test_circuit),
        protocols.unitary(cirq_circuit),
        atol=1e-7,
    )


@pytest.mark.parametrize(
    "common_gate",
    [
        ops.TOFFOLI,
        ops.FREDKIN,
    ],
)
def test_to_from_braket_common_three_qubit_gates(common_gate):
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket -> Cirq.
    """
    cirq_circuit = Circuit(common_gate.on(*LineQubit.range(3)))
    test_circuit = from_braket(to_braket(cirq_circuit))
    testing.assert_allclose_up_to_global_phase(
        protocols.unitary(test_circuit),
        protocols.unitary(cirq_circuit),
        atol=1e-7,
    )

    assert _equal(test_circuit, cirq_circuit, require_qubit_equality=True)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for qiskit executors (qiskit_utils.py)."""

import numpy as np
import pytest
from qiskit import QuantumCircuit
from qiskit_ibm_runtime.fake_provider import FakeLimaV2

from mitiq import MeasurementResult, Observable, PauliString
from mitiq.interface.mitiq_qiskit.qiskit_utils import (
    compute_expectation_value_on_noisy_backend,
    execute,
    execute_with_noise,
    execute_with_shots,
    execute_with_shots_and_noise,
    initialized_depolarizing_noise,
    sample_bitstrings,
)

NOISE = 0.007
ONE_QUBIT_GS_PROJECTOR = np.array([[1, 0], [0, 0]])
TWO_QUBIT_GS_PROJECTOR = np.diag([1, 0, 0, 0])
SHOTS = 1_000


def test_execute():
    """Tests the Qiskit wavefunction simulation executor returns
    appropriate expectation value given an observable.
    """

    circ = QuantumCircuit(1)
    expected_value = execute(circ, obs=ONE_QUBIT_GS_PROJECTOR)
    assert expected_value == 1.0

    second_circ = QuantumCircuit(1)
    second_circ.x(0)
    expected_value = execute(second_circ, obs=ONE_QUBIT_GS_PROJECTOR)
    assert expected_value == 0.0


def test_execute_with_shots():
    """Tests the Qiskit wavefunction sampling simulation executor returns
    appropriate expectation value given an observable.
    """

    circ = QuantumCircuit(1, 1)
    expectation_value = execute_with_shots(
        circuit=circ, obs=ONE_QUBIT_GS_PROJECTOR, shots=SHOTS
    )
    assert expectation_value == 1.0

    second_circ = QuantumCircuit(1)
    second_circ.x(0)
    expectation_value = execute_with_shots(
        circuit=second_circ, obs=ONE_QUBIT_GS_PROJECTOR, shots=SHOTS
    )
    assert expectation_value == 0.0


def test_execute_with_depolarizing_noise_single_qubit():
    """Tests the noisy sampling executor across increasing levels
    of single qubit gate noise
    """

    single_qubit_circ = QuantumCircuit(1)
    # noise model is defined on gates so include the gate to
    # demonstrate noise
    single_qubit_circ.z(0)

    noiseless_exp_value = 1.0

    expectation_value = execute_with_noise(
        circuit=single_qubit_circ,
        obs=ONE_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
    )
    # anticipate that the expectation value will be less than
    # the noiseless simulation of the same circuit
    assert expectation_value < noiseless_exp_value


def test_execute_with_depolarizing_noise_two_qubit():
    """Tests the noisy sampling executor across increasing levels of
    two qubit gate noise.
    """

    two_qubit_circ = QuantumCircuit(2)
    # noise model is defined on gates so include the gate to
    # demonstrate noise
    two_qubit_circ.cx(0, 1)

    noiseless_exp_value = 1.0

    expectation_value = execute_with_noise(
        circuit=two_qubit_circ,
        obs=TWO_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
    )
    # anticipate that the expectation value will be less than
    # the noiseless simulation of the same circuit
    assert expectation_value < noiseless_exp_value


def test_execute_with_shots_and_depolarizing_noise_single_qubit():
    """Tests the noisy sampling executor across increasing levels
    of single qubit gate noise.
    """

    single_qubit_circ = QuantumCircuit(1, 1)
    # noise model is defined on gates so include the gate to
    # demonstrate noise
    single_qubit_circ.z(0)

    noiseless_exp_value = 1.0

    expectation_value = execute_with_shots_and_noise(
        circuit=single_qubit_circ,
        obs=ONE_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
        shots=SHOTS,
    )
    # anticipate that the expectation value will be less than
    # the noiseless simulation of the same circuit
    assert expectation_value < noiseless_exp_value


def test_execute_with_shots_and_depolarizing_noise_two_qubit():
    """Tests the noisy sampling executor across increasing levels of
    two qubit gate noise.
    """

    two_qubit_circ = QuantumCircuit(2, 2)
    # noise model is defined on gates so include the gate to
    # demonstrate noise
    two_qubit_circ.cx(0, 1)

    noiseless_exp_value = 1.0

    expectation_value = execute_with_shots_and_noise(
        circuit=two_qubit_circ,
        obs=TWO_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
        shots=SHOTS,
    )
    # anticipate that the expectation value will be less than
    # the noiseless simulation of the same circuit
    assert expectation_value < noiseless_exp_value


def test_circuit_is_not_mutated_by_executors():
    single_qubit_circ = QuantumCircuit(1, 1)
    single_qubit_circ.z(0)
    expected_circuit = single_qubit_circ.copy()
    execute_with_shots_and_noise(
        circuit=single_qubit_circ,
        obs=ONE_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
        shots=SHOTS,
    )
    assert single_qubit_circ.data == expected_circuit.data
    assert single_qubit_circ == expected_circuit
    execute_with_noise(
        circuit=single_qubit_circ,
        obs=ONE_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(NOISE),
    )
    assert single_qubit_circ.data == expected_circuit.data
    assert single_qubit_circ == expected_circuit


def test_sample_bitstrings():
    """Tests that the function sample_bitstrings returns a valid
    mitiq.MeasurementResult.
    """

    two_qubit_circ = QuantumCircuit(2, 1)
    two_qubit_circ.cx(0, 1)
    two_qubit_circ.measure(0, 0)

    measurement_result = sample_bitstrings(
        circuit=two_qubit_circ,
        backend=None,
        noise_model=initialized_depolarizing_noise(0),
        shots=5,
    )
    assert measurement_result.result == [[0], [0], [0], [0], [0]]
    assert measurement_result.qubit_indices == (0,)


def test_sample_bitstrings_with_measure_all():
    """Tests that the function sample_bitstrings returns a valid
    mitiq.MeasurementResult when "measure_all" is True.
    """
    two_qubit_circ = QuantumCircuit(2)
    two_qubit_circ.cx(0, 1)
    measurement_result = sample_bitstrings(
        circuit=two_qubit_circ,
        backend=None,
        noise_model=initialized_depolarizing_noise(0),
        shots=2,
        measure_all=True,
    )
    assert measurement_result.result == [[0, 0], [0, 0]]
    assert measurement_result.qubit_indices == (0, 1)
    assert isinstance(measurement_result, MeasurementResult)


def test_sample_bitstrings_with_backend():
    """Tests that the function sample_bitstrings returns a valid
    mitiq.MeasurementResult if a qiskit backend is used.
    """
    two_qubit_circ = QuantumCircuit(2)
    two_qubit_circ.cx(0, 1)
    measurement_result = sample_bitstrings(
        circuit=two_qubit_circ,
        backend=FakeLimaV2(),
        shots=5,
        measure_all=True,
    )
    assert len(measurement_result.result) == 5
    assert len(measurement_result.result[0]) == 2
    assert measurement_result.qubit_indices == (0, 1)


def test_sample_bitstrings_error_message():
    """Tests that an error is given backend and nose_model are both None."""
    two_qubit_circ = QuantumCircuit(2)
    two_qubit_circ.cx(0, 1)
    with pytest.raises(ValueError, match="Either a backend or a noise model"):
        sample_bitstrings(
            circuit=two_qubit_circ,
            shots=5,
        )


def test_compute_expectation_value_on_noisy_backend_with_noise_model():
    """Tests the evaluation of an expectation value assuming a noise model."""
    obs = Observable(PauliString("X"))
    qiskit_circuit = QuantumCircuit(1)
    qiskit_circuit.h(0)

    # First we try without noise
    noiseless_expval = compute_expectation_value_on_noisy_backend(
        qiskit_circuit,
        obs,
        noise_model=initialized_depolarizing_noise(0),
    )

    assert isinstance(noiseless_expval, complex)
    assert np.isclose(np.imag(noiseless_expval), 0.0)
    assert np.isclose(np.real(noiseless_expval), 1.0)

    # Now we try with noise
    expval = compute_expectation_value_on_noisy_backend(
        qiskit_circuit,
        obs,
        noise_model=initialized_depolarizing_noise(0.01),
    )

    assert isinstance(expval, complex)
    assert np.isclose(np.imag(expval), 0.0)
    # With noise the result is non-deterministic
    assert 0.9 < np.real(expval) < 1.0


def test_compute_expectation_value_on_noisy_backend_with_qiskit_backend():
    """Tests the evaluation of an expectation value on a noisy backed"""
    obs = Observable(PauliString("X"))
    qiskit_circuit = QuantumCircuit(1)
    qiskit_circuit.h(0)

    expval = compute_expectation_value_on_noisy_backend(
        qiskit_circuit,
        obs,
        backend=FakeLimaV2(),
    )

    assert isinstance(expval, complex)
    assert np.isclose(np.imag(expval), 0.0)
    # With noise the result is non-deterministic
    assert 0.9 < np.real(expval) < 1.0



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests to make sure Mitiq works without optional packages like Qiskit,
pyQuil, etc.

Ideally these tests should touch all of Mitiq except for
mitiq.interface.mitiq_[package], where [package] is any supported package that
interfaces with Mitiq (see mitiq.SUPPORTED_PROGRAM_TYPES).
"""

from typing import Union, get_origin

from cirq import Circuit


def test_import():
    """Simple test that Mitiq can be imported without any (or all) supported
    program types.
    """
    import mitiq

    if get_origin(mitiq.QPROGRAM) is Union:
        assert (
            1  # cirq.Circuit is always supported.
            <= len(mitiq.QPROGRAM.__args__)  # All installed types.
            <= len(mitiq.SUPPORTED_PROGRAM_TYPES.keys())  # All types.
        )
    else:
        assert mitiq.QPROGRAM is Circuit


# TODO: More tests wanted!



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for the Quantum Subspace Expansion top level API."""

from typing import List
from unittest.mock import patch

import cirq
import numpy as np
import pytest

from mitiq import QPROGRAM, Observable, PauliString
from mitiq.interface import convert_to_mitiq
from mitiq.interface.mitiq_cirq import compute_density_matrix
from mitiq.qse import (
    execute_with_qse,
    get_projector,
    mitigate_executor,
    qse_decorator,
)
from mitiq.qse.qse_utils import _compute_overlap_matrix


def execute_with_depolarized_noise(circuit: QPROGRAM) -> np.ndarray:
    return compute_density_matrix(
        convert_to_mitiq(circuit)[0],
        noise_model_function=cirq.depolarize,
        noise_level=(0.01,),
    )


def batched_execute_with_depolarized_noise(circuits) -> List[np.ndarray]:
    return [execute_with_depolarized_noise(circuit) for circuit in circuits]


def execute_no_noise(circuit: QPROGRAM) -> np.ndarray:
    return compute_density_matrix(
        convert_to_mitiq(circuit)[0], noise_level=(0,)
    )


@pytest.fixture
def prepare_setup():
    qc = prepare_logical_0_state_for_5_1_3_code()
    (
        check_operators,
        code_hamiltonian,
    ) = get_5_1_3_code_check_operators_and_code_hamiltonian()
    return qc, check_operators, code_hamiltonian


def test_execute_with_qse_no_noise(prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    observable = get_observable_in_code_space(PauliString("ZZZZZ"))
    mitigated_value = execute_with_qse(
        qc, execute_no_noise, check_operators, code_hamiltonian, observable
    )
    assert np.isclose(mitigated_value.real, 1)

    observable = get_observable_in_code_space(PauliString("XXXXX"))
    mitigated_value = execute_with_qse(
        qc, execute_no_noise, check_operators, code_hamiltonian, observable
    )
    assert np.isclose(mitigated_value.real, 0.5)


def test_execute_with_qse(prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    observable = get_observable_in_code_space(PauliString("ZZZZZ"))
    unmitigated_value = observable.expectation(
        qc, execute_with_depolarized_noise
    )
    mitigated_value = execute_with_qse(
        qc,
        execute_with_depolarized_noise,
        check_operators,
        code_hamiltonian,
        observable,
    )
    assert abs(mitigated_value.real - 1) < abs(unmitigated_value.real - 1)

    observable = get_observable_in_code_space(PauliString("XXXXX"))
    unmitigated_value = observable.expectation(
        qc, execute_with_depolarized_noise
    )
    mitigated_value = execute_with_qse(
        qc,
        execute_with_depolarized_noise,
        check_operators,
        code_hamiltonian,
        observable,
    )
    assert abs(mitigated_value.real - 0.5) < abs(unmitigated_value.real - 0.5)


@patch("mitiq.qse.qse.execute_with_qse")
def test_mitigate_executor_batched(mock_execute_with_qse, prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    observable = get_observable_in_code_space(PauliString("XXXXX"))
    batched_mitigated_executor = mitigate_executor(
        batched_execute_with_depolarized_noise,
        check_operators,
        code_hamiltonian,
        observable,
    )

    num_circuits = 3
    circuits = [qc] * num_circuits
    batched_mitigated_executor(circuits)

    mock_execute_with_qse.assert_called_with(
        circuits[0],
        batched_execute_with_depolarized_noise,
        check_operators,
        code_hamiltonian,
        observable,
        {},
    )
    assert mock_execute_with_qse.call_count == num_circuits


@patch("mitiq.qse.qse.execute_with_qse")
def test_qse_decorator(mock_execute_with_qse, prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    observable = get_observable_in_code_space(PauliString("ZZZZZ"))

    @qse_decorator(
        check_operators=check_operators,
        code_hamiltonian=code_hamiltonian,
        observable=observable,
    )
    def decorated_executor(circuit: QPROGRAM) -> np.ndarray:
        return compute_density_matrix(
            convert_to_mitiq(circuit)[0],
            noise_model_function=cirq.depolarize,
            noise_level=(0.01,),
        )

    decorated_executor(qc)
    mock_execute_with_qse.assert_called_once()


def test_get_projector(prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    P = get_projector(qc, execute_no_noise, check_operators, code_hamiltonian)
    uniform_projector = Observable(*[-0.25 * c for c in check_operators])
    assert P == uniform_projector


def test_compute_overlap_matrix(prepare_setup):
    qc, check_operators, _ = prepare_setup

    S = _compute_overlap_matrix(qc, execute_no_noise, check_operators, {})
    assert np.allclose(S, np.ones(16))

    S = _compute_overlap_matrix(
        qc, execute_with_depolarized_noise, check_operators, {}
    )
    # Diagonal terms are all 1 because
    # ⟨Ψ|C_i C_i|Ψ⟩ = ⟨Ψ|Ψ⟩ = 1
    assert np.allclose(np.diag(np.diag(S)), np.eye(16))

    # Off diagonal terms are less than 1 because
    # ⟨Ψ|C_i C_j|Ψ⟩  = ⟨Ψ|C_k|Ψ⟩ < 1
    # (since |Ψ⟩ was rotated a bit from the logical subspace)
    # All off diagonal terms are the same because of the symmetry of the
    # total depolarizing noise.
    off_diag_elements = S[np.where(~np.eye(16, dtype=bool))]
    assert np.allclose(off_diag_elements, off_diag_elements[0])
    assert off_diag_elements[0] < 1


def test_compute_overlap_matrix_with_hamiltonian(prepare_setup):
    qc, check_operators, code_hamiltonian = prepare_setup

    # If we have a full set of check operators that form a group then all
    # entries of the H matrix should be the same.
    # H_jk = sum over all i's ⟨Ψ|C_i|Ψ⟩

    H = _compute_overlap_matrix(
        qc, execute_no_noise, check_operators, {}, code_hamiltonian
    )
    assert np.allclose(H, np.full(16, -16))

    H = _compute_overlap_matrix(
        qc,
        execute_with_depolarized_noise,
        check_operators,
        {},
        code_hamiltonian,
    )
    assert np.allclose(H, H[0][0])
    assert H[0][0].real > -16


def get_observable_in_code_space(observable: cirq.PauliString):
    FIVE_I = PauliString("IIIII")
    projector_onto_code_space = [
        PauliString("XZZXI"),
        PauliString("IXZZX"),
        PauliString("XIXZZ"),
        PauliString("ZXIXZ"),
    ]

    observable_in_code_space = Observable(FIVE_I)
    all_paulis = projector_onto_code_space + [observable]
    for g in all_paulis:
        observable_in_code_space *= 0.5 * Observable(FIVE_I, g)
    return observable_in_code_space


def get_5_1_3_code_check_operators_and_code_hamiltonian() -> tuple:
    """
    Returns the check operators and code Hamiltonian for the [[5,1,3]] code
    The check operators are computed from the stabilizer generators:
    (1+G1)(1+G2)(1+G3)(1+G4)  G = [XZZXI, IXZZX, XIXZZ, ZXIXZ]
    source: https://en.wikipedia.org/wiki/Five-qubit_error_correcting_code
    """
    Ms = [
        "YIYXX",
        "ZIZYY",
        "IXZZX",
        "ZXIXZ",
        "YYZIZ",
        "XYIYX",
        "YZIZY",
        "ZZXIX",
        "XZZXI",
        "ZYYZI",
        "IYXXY",
        "IZYYZ",
        "YXXYI",
        "XXYIY",
        "XIXZZ",
        "IIIII",
    ]
    Ms_as_pauliStrings = [
        PauliString(M, coeff=1, support=range(5)) for M in Ms
    ]
    negative_Ms_as_pauliStrings = [
        PauliString(M, coeff=-1, support=range(5)) for M in Ms
    ]
    Hc = Observable(*negative_Ms_as_pauliStrings)
    return Ms_as_pauliStrings, Hc


def prepare_logical_0_state_for_5_1_3_code():
    """
    To simplify the testing logic. We hardcode the the logical 0 and logical 1
    states of the [[5,1,3]] code, copied from:
    https://en.wikipedia.org/wiki/Five-qubit_error_correcting_code
    We then use Gram-Schmidt orthogonalization to fill up the rest of the
    matrix with orthonormal vectors.
    Following this we construct a circuit that has this matrix as its gate.
    """

    def gram_schmidt(
        orthogonal_vecs: List[np.ndarray],
    ) -> np.ndarray:
        orthonormalVecs = [
            vec / np.sqrt(np.vdot(vec, vec)) for vec in orthogonal_vecs
        ]
        dim = np.shape(orthogonal_vecs[0])[0]
        for i in range(dim - len(orthogonal_vecs)):
            new_vec = np.zeros(dim)
            new_vec[i] = 1
            projs = sum(
                [
                    np.vdot(new_vec, cached_vec) * cached_vec
                    for cached_vec in orthonormalVecs
                ]
            )
            new_vec -= projs
            orthonormalVecs.append(
                new_vec / np.sqrt(np.vdot(new_vec, new_vec))
            )
        return np.reshape(orthonormalVecs, (32, 32)).T

    logical_0_state = np.zeros(32)

    # bitstrings ["00000", "10010", "01001", "10100", "01010", "00101"]
    np.put(logical_0_state, [0, 18, 9, 20, 10, 5], 1 / 4)

    # bitstrings: [
    #  "11011", "00110", "11000", "11101", "00011",
    #  "11110", "01111", "10001", "01100", "10111"
    # ]
    np.put(logical_0_state, [27, 6, 24, 29, 3, 30, 15, 17, 12, 23], -1 / 4)

    logical_1_state = np.zeros(32)

    # bitstrings ["11111", "01101", "10110", "01011", "10101", "11010"]
    np.put(logical_1_state, [31, 13, 22, 11, 21, 26], 1 / 4)

    # bitstrings: [
    #  "00100", "11001", "00111", "00010", "11100",
    #  "00001", "10000", "01110", "10011", "01000"
    # ]
    np.put(logical_1_state, [4, 25, 7, 2, 28, 1, 16, 14, 19, 8], -1 / 4)

    # Fill up the rest of the matrix with orthonormal vectors
    matrix = gram_schmidt([logical_0_state, logical_1_state])
    circuit = cirq.Circuit()
    g = cirq.MatrixGate(matrix)
    qubits = cirq.LineQubit.range(5)
    circuit.append(g(*qubits))

    return circuit



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for generating (near) Clifford circuits."""

import cirq
import numpy as np
import pytest
from cirq.circuits import Circuit

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.cdr.clifford_utils import (
    _CLIFFORD_ANGLES,
    angle_to_proximities,
    angle_to_proximity,
    closest_clifford,
    count_non_cliffords,
    is_clifford,
    is_clifford_angle,
    probabilistic_angle_to_clifford,
    random_clifford,
)
from mitiq.interface import convert_from_mitiq


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_is_clifford_with_clifford(circuit_type):
    circuit = convert_from_mitiq(
        cirq.Circuit(cirq.Z.on(cirq.LineQubit(0))), circuit_type
    )
    assert is_clifford(circuit)


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_is_clifford_with_nonclifford(circuit_type):
    circuit = convert_from_mitiq(
        cirq.Circuit(cirq.T.on(cirq.LineQubit(0))), circuit_type
    )
    assert not is_clifford(circuit)


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_count_non_cliffords(circuit_type):
    a, b = cirq.LineQubit.range(2)
    circuit = Circuit(
        cirq.rz(0.0).on(a),  # Clifford.
        cirq.rx(0.1 * np.pi).on(b),  # Non-Clifford.
        cirq.rx(0.5 * np.pi).on(b),  # Clifford
        cirq.rz(0.4 * np.pi).on(b),  # Non-Clifford.
        cirq.rz(0.5 * np.pi).on(b),  # Clifford.
        cirq.CNOT.on(a, b),  # Clifford.
    )
    circuit = convert_from_mitiq(circuit, circuit_type)

    assert count_non_cliffords(circuit) == 2


def test_count_non_cliffords_empty_circuit():
    assert count_non_cliffords(Circuit()) == 0


def test_is_clifford_angle():
    for p in range(4):
        assert is_clifford_angle(p * np.array(_CLIFFORD_ANGLES)).all()

    assert not is_clifford_angle(-0.17)


def test_closest_clifford():
    for ang in _CLIFFORD_ANGLES:
        angs = np.linspace(ang - np.pi / 4 + 0.01, ang + np.pi / 4 - 0.01)
        for a in angs:
            assert closest_clifford(a) == ang


def test_random_clifford():
    assert set(random_clifford(20, np.random.RandomState(1))).issubset(
        _CLIFFORD_ANGLES
    )


def test_angle_to_proximities():
    for sigma in np.linspace(0.1, 2, 10):
        for ang in _CLIFFORD_ANGLES:
            probabilities = angle_to_proximities(ang, sigma)
            assert (isinstance(p, float) for p in probabilities)


def test_angle_to_proximity():
    for sigma in np.linspace(0.1, 2, 10):
        probabilities = angle_to_proximity(_CLIFFORD_ANGLES, sigma)
        assert all(isinstance(p, float) for p in probabilities)


def test_probabilistic_angles_to_clifford():
    for sigma in np.linspace(0.1, 2, 10):
        angles = probabilistic_angle_to_clifford(
            _CLIFFORD_ANGLES, sigma, np.random.RandomState(1)
        )
        assert all(a in _CLIFFORD_ANGLES for a in angles)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""NB: Copied in large part from rigetti/forest-benchmarking (Apache-2.0)
and modified to test a larger gateset.
"""

import inspect
import itertools
import random
from math import pi

import numpy as np
import pytest
from cirq import equal_up_to_global_phase
from pyquil.gates import (
    CCNOT,
    CNOT,
    CPHASE,
    CSWAP,
    CZ,
    ISWAP,
    PHASE,
    RX,
    RY,
    RZ,
    SWAP,
    XY,
    H,
    I,
    S,
    T,
    X,
    Y,
    Z,
)
from pyquil.quil import Program
from pyquil.simulation.tools import program_unitary

from mitiq.interface.mitiq_pyquil.compiler import (
    _CCNOT,
    _CNOT,
    _CPHASE,
    _H,
    _ISWAP,
    _RY,
    _S,
    _SWAP,
    _T,
    _X,
    _Y,
    _Z,
    basic_compile,
)


def test_basic_compile_defgate():
    p = Program()
    p.inst(RX(pi, 0))
    p.defgate("test", [[0, 1], [1, 0]])
    p.inst(("test", 2))
    p.inst(RZ(pi / 2, 0))

    assert p == basic_compile(p)


def test_CCNOT():
    for perm in itertools.permutations([0, 1, 2]):
        u1 = program_unitary(Program(CCNOT(*perm)), n_qubits=3)
        u2 = program_unitary(_CCNOT(*perm), n_qubits=3)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_CNOT():
    u1 = program_unitary(Program(CNOT(0, 1)), n_qubits=2)
    u2 = program_unitary(_CNOT(0, 1), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)

    u1 = program_unitary(Program(CNOT(1, 0)), n_qubits=2)
    u2 = program_unitary(_CNOT(1, 0), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_CPHASE():
    for theta in np.linspace(-2 * np.pi, 2 * np.pi):
        u1 = program_unitary(Program(CPHASE(theta, 0, 1)), n_qubits=2)
        u2 = program_unitary(_CPHASE(theta, 0, 1), n_qubits=2)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)

        u1 = program_unitary(Program(CPHASE(theta, 1, 0)), n_qubits=2)
        u2 = program_unitary(_CPHASE(theta, 1, 0), n_qubits=2)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_H():
    u1 = program_unitary(Program(H(0)), n_qubits=1)
    u2 = program_unitary(_H(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_ISWAP():
    u1 = program_unitary(Program(ISWAP(0, 1)), n_qubits=2)
    u2 = program_unitary(_ISWAP(0, 1), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)

    u1 = program_unitary(Program(ISWAP(1, 0)), n_qubits=2)
    u2 = program_unitary(_ISWAP(1, 0), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_RX():
    for theta in np.linspace(-2 * np.pi, 2 * np.pi):
        p = Program(RX(theta, 0))
        u1 = program_unitary(p, n_qubits=1)
        u2 = program_unitary(basic_compile(p), n_qubits=1)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_RY():
    for theta in np.linspace(-2 * np.pi, 2 * np.pi):
        u1 = program_unitary(Program(RY(theta, 0)), n_qubits=1)
        u2 = program_unitary(_RY(theta, 0), n_qubits=1)
        assert equal_up_to_global_phase(u1, u2)


def test_S():
    u1 = program_unitary(Program(S(0)), n_qubits=1)
    u2 = program_unitary(_S(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_SWAP():
    u1 = program_unitary(Program(SWAP(0, 1)), n_qubits=2)
    u2 = program_unitary(_SWAP(0, 1), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)

    u1 = program_unitary(Program(SWAP(1, 0)), n_qubits=2)
    u2 = program_unitary(_SWAP(1, 0), n_qubits=2)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_T():
    u1 = program_unitary(Program(T(0)), n_qubits=1)
    u2 = program_unitary(_T(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_X():
    u1 = program_unitary(Program(X(0)), n_qubits=1)
    u2 = program_unitary(_X(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_XY():
    for theta in np.linspace(-2 * np.pi, 2 * np.pi):
        p = Program(XY(theta, 0, 1))
        u1 = program_unitary(p, n_qubits=2)
        u2 = program_unitary(basic_compile(p), n_qubits=2)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)

        p = Program(XY(theta, 1, 0))
        u1 = program_unitary(p, n_qubits=2)
        u2 = program_unitary(basic_compile(p), n_qubits=2)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_Y():
    u1 = program_unitary(Program(Y(0)), n_qubits=1)
    u2 = program_unitary(_Y(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_Z():
    u1 = program_unitary(Program(Z(0)), n_qubits=1)
    u2 = program_unitary(_Z(0), n_qubits=1)
    assert equal_up_to_global_phase(u1, u2, atol=1e-12)


# Note to developers: unsupported gates are commented out.
QUANTUM_GATES = {
    "I": I,
    "X": X,
    "Y": Y,
    "Z": Z,
    "H": H,
    "S": S,
    "T": T,
    "PHASE": PHASE,
    "RX": RX,
    "RY": RY,
    "RZ": RZ,
    "CZ": CZ,
    "CNOT": CNOT,
    "CCNOT": CCNOT,
    # 'CPHASE00': CPHASE00,
    # 'CPHASE01': CPHASE01,
    # 'CPHASE10': CPHASE10,
    "CPHASE": CPHASE,
    "SWAP": SWAP,
    # 'CSWAP': CSWAP,
    "ISWAP": ISWAP,
    # 'PSWAP': PSWAP
}


def _generate_random_program(n_qubits, length):
    """Randomly sample gates and arguments (qubits, angles)"""
    if n_qubits < 3:
        raise ValueError(
            "Please request n_qubits >= 3 so we can use 3-qubit gates."
        )

    gates = list(QUANTUM_GATES.values())
    prog = Program()
    for _ in range(length):
        gate = random.choice(gates)
        possible_qubits = set(range(n_qubits))
        sig = inspect.signature(gate)

        param_vals = []
        for param in sig.parameters:
            if param in [
                "qubit",
                "q1",
                "q2",
                "control",
                "control1",
                "control2",
                "target",
                "target_1",
                "target_2",
                "classical_reg",
            ]:
                param_val = random.choice(list(possible_qubits))
                possible_qubits.remove(param_val)
            elif param == "angle":
                # TODO: support rx(theta)
                if gate == RX:
                    param_val = random.choice([-1, -0.5, 0, 0.5, 1]) * pi
                else:
                    param_val = random.uniform(-2 * pi, 2 * pi)
            else:
                raise ValueError("Unknown gate parameter {}".format(param))

            param_vals.append(param_val)

        prog += gate(*param_vals)

    return prog


@pytest.fixture(params=list(range(3, 5)))
def n_qubits(request):
    return request.param


@pytest.fixture(params=[2, 50, 67])
def prog_length(request):
    return request.param


def test_random_progs(n_qubits, prog_length):
    for repeat_i in range(10):
        prog = _generate_random_program(n_qubits=n_qubits, length=prog_length)
        u1 = program_unitary(prog, n_qubits=n_qubits)
        u2 = program_unitary(basic_compile(prog), n_qubits=n_qubits)
        assert equal_up_to_global_phase(u1, u2, atol=1e-12)


def test_unsupported_gate():
    with pytest.raises(ValueError):
        basic_compile(Program(CSWAP(0, 1, 2)))


def test_other_instructions():
    p = Program("DECLARE ro BIT[2]")
    assert p == basic_compile(p)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Reads in version information."""

import os
from importlib.metadata import PackageNotFoundError, version

directory_of_this_file = os.path.dirname(os.path.abspath(__file__))

try:
    __version__ = version("mitiq")
except PackageNotFoundError:
    directory_of_this_file = os.path.dirname(os.path.abspath(__file__))
    with open(f"{directory_of_this_file}/../VERSION.txt", "r") as f:
        __version__ = f.read().strip()



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest
from cirq import (
    CCNOT,
    CNOT,
    CZ,
    ISWAP,
    SWAP,
    Circuit,
    Gate,
    H,
    I,
    LineQubit,
    X,
    Y,
    Z,
    ops,
    unitary,
)

from mitiq.pec.channels import _circuit_to_choi, _operation_to_choi
from mitiq.pec.representations.biased_noise import (
    represent_operation_with_local_biased_noise,
)


def single_qubit_biased_noise_overhead(epsilon: float, eta: float) -> float:
    """Overhead calculation similar to that presented in
    :cite:`Temme_2017_PRL` and modified according to combined (biased) noise
    channel in :cite:`Strikis_2021_PRXQuantum`.
    """
    a = 1 - epsilon
    b = epsilon * (3 * eta + 1) / (3 * (eta + 1))
    c = epsilon / (3 * (eta + 1))
    eta1 = (a**2 + a * b - 2 * c**2) / (
        a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2
    )
    eta2 = -c / (a**2 + 2 * a * b + b**2 - 4 * c**2)
    eta3 = (-a * b - b**2 + 2 * c**2) / (
        a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2
    )

    return abs(eta1) + 2 * abs(eta2) + abs(eta3)


def two_qubit_biased_noise_overhead(epsilon: float, eta: float) -> float:
    """Overhead calculation similar to that presented in
    :cite:`Temme_2017_PRL` and modified according to combined (biased) noise
    channel in :cite:`Strikis_2021_PRXQuantum`.
    """
    a = 1 - epsilon
    b = epsilon * (3 * eta + 1) / (3 * (eta + 1))
    c = epsilon / (3 * (eta + 1))
    alpha_sq = (a**2 + a * b - 2 * c**2) ** 2 / (
        a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2
    ) ** 2
    alpha_beta = (
        (a**2 + a * b - 2 * c**2)
        * (-a * b - b**2 + 2 * c**2)
        / (a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2)
        ** 2
    )
    alpha_gamma = (
        -c
        * (a**2 + a * b - 2 * c**2)
        / (
            (a**2 + 2 * a * b + b**2 - 4 * c**2)
            * (a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2)
        )
    )
    beta_sq = (-a * b - b**2 + 2 * c**2) ** 2 / (
        a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2
    ) ** 2
    beta_gamma = (
        -c
        * (-a * b - b**2 + 2 * c**2)
        / (
            (a**2 + 2 * a * b + b**2 - 4 * c**2)
            * (a**3 + a**2 * b - a * b**2 - 4 * a * c**2 - b**3 + 4 * b * c**2)
        )
    )
    gamma_sq = c**2 / (a**2 + 2 * a * b + b**2 - 4 * c**2) ** 2
    overhead = (
        alpha_sq
        + 4 * abs(alpha_gamma)
        + 2 * abs(alpha_beta)
        + 4 * (gamma_sq)
        + 4 * abs(beta_gamma)
        + beta_sq
    )

    return overhead


@pytest.mark.parametrize("epsilon", [0, 0.1, 0.7])
@pytest.mark.parametrize("eta", [0, 1, 1000])
@pytest.mark.parametrize("gate", [X, Y, Z, H])
def test_single_qubit_representation_norm(
    gate: Gate, epsilon: float, eta: float
):
    q = LineQubit(0)
    optimal_norm = single_qubit_biased_noise_overhead(epsilon, eta)
    norm = represent_operation_with_local_biased_noise(
        Circuit(gate(q)), epsilon, eta
    ).norm
    assert np.isclose(optimal_norm, norm)


@pytest.mark.parametrize("epsilon", [0, 0.1, 0.7])
@pytest.mark.parametrize("eta", [0, 1, 1000])
@pytest.mark.parametrize("gate", (CZ, CNOT, ISWAP, SWAP))
def test_two_qubit_representation_norm(gate: Gate, epsilon: float, eta: float):
    qreg = LineQubit.range(2)
    optimal_norm = two_qubit_biased_noise_overhead(epsilon, eta)
    norm = represent_operation_with_local_biased_noise(
        Circuit(gate(*qreg)), epsilon, eta
    ).norm
    assert np.isclose(optimal_norm, norm)


def test_three_qubit_biased_noise_representation_error():
    q0, q1, q2 = LineQubit.range(3)
    with pytest.raises(ValueError):
        represent_operation_with_local_biased_noise(
            Circuit(CCNOT(q0, q1, q2)), 0.05, 10
        )


@pytest.mark.parametrize("epsilon", [0, 0.1, 0.7])
@pytest.mark.parametrize("eta", [0, 1, 1000])
@pytest.mark.parametrize("gate", [X, Y, Z, H, CZ, CNOT, ISWAP, SWAP])
def test_biased_noise_representation_with_choi(
    gate: Gate, epsilon: float, eta: float
):
    """Tests the representation by comparing exact Choi matrices."""
    qreg = LineQubit.range(gate.num_qubits())
    ideal_choi = _operation_to_choi(gate.on(*qreg))
    op_rep = represent_operation_with_local_biased_noise(
        Circuit(gate.on(*qreg)), epsilon, eta
    )
    choi_components = []

    # Define biased noise channel
    a = 1 - epsilon
    b = epsilon * (3 * eta + 1) / (3 * (eta + 1))
    c = epsilon / (3 * (eta + 1))

    mix = [
        (a, unitary(I)),
        (b, unitary(Z)),
        (c, unitary(X)),
        (c, unitary(Y)),
    ]

    for coeff, noisy_op in op_rep.basis_expansion:
        implementable_circ = noisy_op.circuit
        # Apply noise after each sequence.
        # NOTE: noise is not applied after each operation.
        biased_op = ops.MixedUnitaryChannel(mix).on_each(*qreg)
        implementable_circ.append(biased_op)
        sequence_choi = _circuit_to_choi(implementable_circ)
        choi_components.append(coeff * sequence_choi)
    combination_choi = np.sum(choi_components, axis=0)
    assert np.allclose(ideal_choi, combination_choi, atol=10**-6)
